/*
 *          File: HartFrame.cpp (CFrame)
 *                The hart frame is a construct used to collect all information
 *                which is needed to encode and decode data of so called
 *                service primitives like responses and requests, which are
 *                finally octet streams.
 *
 *        Author: Walter Borst
 *
 *        E-Mail: info@borst-automation.de
 *          Home: https://www.borst-automation.de
 *
 * No Warranties: https://www.borst-automation.com/legal/warranty-disclaimer
 *
 * Copyright 2006-2024 Walter Borst, Cuxhaven, Germany
 */

#include "OSAL.h"
#include "WbHartUser.h"
#include "HartFrame.h"
#include "HartLib.h"

// Initialization
void CFrame::Init()
{
    ClearBuffers();
    m_num_preambles = 0;
    m_status = EN_Status::GET_PREAMBLE;
    m_start_time = 0;
    m_burst_frame = EN_Bool::FALSE8;
    m_rcv_byte_count = 0;
    m_payload_count = 0;
    m_exp_bytes_count = 0;
    m_payload_size = 0;
    m_active = EN_Bool::TRUE8;
}

void CFrame::Uninit()
{
    m_active = EN_Bool::FALSE8;
}

EN_Bool CFrame::IsActive()
{
    return m_active;
}

void CFrame::SetActive()
{
    m_active = EN_Bool::TRUE8;
}

void CFrame::ClearActive()
{
    m_active = EN_Bool::FALSE8;
}

// Online Parsing
EN_Bool CFrame::TryParse(TY_Word* bytes_parsed_, TY_Byte* new_data_, TY_Byte* new_err_, TY_Word new_data_len_, EN_Bool gap_time_out_)
{
    TY_Word bytes_parsed = 0;
    //static TY_Byte  test[256];

    // Just for testing copy data to local buffer
    //COSAL::CMem::Set(test, 0 , 256);
    //COSAL::CMem::Copy(test, new_data_, new_data_len_ );

    while ((bytes_parsed < new_data_len_) &&
        (m_status != EN_Status::FRAME_COMPLETED) &&
        (m_status != EN_Status::INVALID_FRAME) &&
        (m_status != EN_Status::CHECKSUM_ERR)
        )
    {
        switch (m_status)
        {
        // State Machine Cases
        case EN_Status::GET_PREAMBLE:
            m_status = GetPreamble(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GETTING_PREAMBLES:
            m_status = GettingPreambles(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_DELIMITER:
            m_status = GetDelimiter(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_ADDRESS:
            m_status = GetAddress(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GETTING_LONG_ADDR:
            m_status = GettingLongAddress(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_EXP_BYTES:
            m_status = GetExpBytes(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GETTING_EXP_BYTES:
            m_status = GettingExpBytes(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_CMD:
            m_status = GetCmd(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_DATA_LEN:
            m_status = GetDataLen(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_RESP1:
            m_status = GetResponse1(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_RESP2:
            m_status = GetResponse2(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_PAYLOAD:
            m_status = GetPayload(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GETTING_PAYLOAD:
            m_status = GettingPayload(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_CHECKSUM:
            m_status = GetCheckSum(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GET_GARBAGE:
            m_status = GetGarbage(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        case EN_Status::GETTING_GARBAGE:
            m_status = GettingGarbage(new_data_[bytes_parsed], new_err_[bytes_parsed]);
            break;
        default:
            break;
        }
        bytes_parsed++;
    }
    if (gap_time_out_ == EN_Bool::TRUE8)
    {
        if (m_status != EN_Status::GET_PREAMBLE)
        {
            // handle gap time out
            bytes_parsed = new_data_len_;
            m_status = EN_Status::INVALID_FRAME;
        }
        else
        {
            if (m_rcv_byte_count > 0)
            {
                bytes_parsed = new_data_len_;
                m_status = EN_Status::INVALID_FRAME;
            }
        }
    }
    
    *bytes_parsed_ = bytes_parsed;
    m_rcv_byte_count += bytes_parsed;
    
    if ((m_status == EN_Status::CHECKSUM_ERR) ||
        (m_status == EN_Status::INVALID_FRAME) ||
        (m_status == EN_Status::FRAME_COMPLETED)
        )
    {
        return EN_Bool::TRUE8;
    }
    return EN_Bool::FALSE8;
}

// Parsing State Machine Functions
CFrame::EN_Status CFrame::GetPreamble(TY_Byte data_, TY_Byte error_)
{
    if (IsPreamble(data_) == EN_Bool::TRUE8)
    {
        m_num_preambles = 1;
        return EN_Status::GETTING_PREAMBLES;
    }
    return EN_Status::GET_PREAMBLE;
}

CFrame::EN_Status CFrame::GettingPreambles(TY_Byte data_, TY_Byte error_)
{
    if (IsPreamble(data_) == EN_Bool::TRUE8)
    {
        m_num_preambles = 2;
        return EN_Status::GET_DELIMITER;
    }
    return EN_Status::GET_PREAMBLE;
}

CFrame::EN_Status CFrame::GetDelimiter(TY_Byte data_, TY_Byte error_)
{
    if (IsPreamble(data_) == EN_Bool::TRUE8)
    {
        m_num_preambles++;
        if (m_num_preambles > CHart::CLimit::MAX_NUM_PREAMBLES)
        {
            return EN_Status::GET_GARBAGE;
        }
        else
        {
            return EN_Status::GET_DELIMITER;
        }
    }
    else if (IsDelimiter(data_) == EN_Bool::TRUE8)
    {
        m_num_exp_bytes = (TY_Byte)((data_ & CHart::CMask::NUM_EXP_BYTES) >> 5);
        m_delimiter = data_ & CHart::CMask::DELIMITER;
        m_target_chk = data_; // Start checksum
        if (IsLongAddress(data_) == EN_Bool::TRUE8)
        {
            m_addr_type = CHart::CAddrType::LONG;
        }
        else
        {
            m_addr_type = CHart::CAddrType::SHORT;
        }
        return EN_Status::GET_ADDRESS;
    }
    return EN_Status::GET_PREAMBLE;
}

CFrame::EN_Status CFrame::GetAddress(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    m_address[0] = data_ & CHart::CMask::HIGH_ADDR;
    if (IsPrimaryMaster(data_) == EN_Bool::TRUE8)
    {
        m_remote_master = EN_Bool::TRUE8;
    }
    else
    {
        m_remote_master = EN_Bool::FALSE8;
    }
    if (IsBurstModeFlag(data_) == EN_Bool::TRUE8)
    {
        m_burst_mode = EN_Bool::TRUE8;
    }
    else
    {
        m_burst_mode = EN_Bool::FALSE8;
    }
    m_addr_byte_count = 1;
    if (m_addr_type == CHart::CAddrType::LONG)
    {
        return EN_Status::GETTING_LONG_ADDR;
    }
    if (m_num_exp_bytes > 0)
    {
        return EN_Status::GET_EXP_BYTES;
    }
    return EN_Status::GET_CMD;
}

CFrame::EN_Status CFrame::GettingLongAddress(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    m_address[m_addr_byte_count] = data_;
    m_addr_byte_count++;
    if (m_addr_byte_count == CHart::CSize::NUM_ADDR_BYTES)
    {
        if (m_num_exp_bytes > 0)
        {
            return EN_Status::GET_EXP_BYTES;
        }
        return EN_Status::GET_CMD;
    }
    return EN_Status::GETTING_LONG_ADDR;
}

CFrame::EN_Status CFrame::GetExpBytes(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    m_exp_bytes[0] = data_;
    m_exp_bytes_count = 1;
    if (m_num_exp_bytes == 1)
    {
        return EN_Status::GET_CMD;
    }
    return EN_Status::GETTING_EXP_BYTES;
}

CFrame::EN_Status CFrame::GettingExpBytes(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    m_exp_bytes[m_exp_bytes_count] = data_;
    m_exp_bytes_count++;
    if (m_exp_bytes_count == m_num_exp_bytes)
    {
        return EN_Status::GET_CMD;
    }
    return EN_Status::GETTING_EXP_BYTES;
}

CFrame::EN_Status CFrame::GetCmd(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    m_cmd = data_;
    return EN_Status::GET_DATA_LEN;
}

CFrame::EN_Status CFrame::GetDataLen(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    if (m_delimiter == CHart::CDelimiter::STX)
    {
        m_payload_size = data_;
        if (m_payload_size > 0)
        {
            return EN_Status::GET_PAYLOAD;
        }
        else
        {
            return EN_Status::GET_CHECKSUM;
        }
    }
    else
    {
        if (data_ >= 2)
        {
            m_payload_size = data_ - 2;
            return EN_Status::GET_RESP1;
        }
    }
    return EN_Status::INVALID_FRAME;
}

CFrame::EN_Status CFrame::GetResponse1(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    m_rsp1 = data_;
    return EN_Status::GET_RESP2;
}

CFrame::EN_Status CFrame::GetResponse2(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    m_rsp2 = data_;
    if (m_payload_size > 0)
    {
        return EN_Status::GET_PAYLOAD;
    }
    else
    {
        return EN_Status::GET_CHECKSUM;
    }
}

CFrame::EN_Status CFrame::GetPayload(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    m_payload_data[0] = data_;
    m_payload_count = 1;
    if (m_payload_size == 1)
    {
        return EN_Status::GET_CHECKSUM;
    }
    return EN_Status::GETTING_PAYLOAD;
}

CFrame::EN_Status CFrame::GettingPayload(TY_Byte data_, TY_Byte error_)
{
    m_target_chk ^= data_;
    m_payload_data[m_payload_count] = data_;
    m_payload_count++;
    if (m_payload_count == m_payload_size)
    {
        return EN_Status::GET_CHECKSUM;
    }
    return EN_Status::GETTING_PAYLOAD;
}

CFrame::EN_Status CFrame::GetCheckSum(TY_Byte data_, TY_Byte error_)
{
    m_actual_chk = data_;
    if (m_actual_chk == m_target_chk)
    {
        return EN_Status::FRAME_COMPLETED;
    }
    return EN_Status::CHECKSUM_ERR;
}

CFrame::EN_Status CFrame::GetGarbage(TY_Byte data_, TY_Byte error_)
{
    return EN_Status::GETTING_GARBAGE;
}

CFrame::EN_Status CFrame::GettingGarbage(TY_Byte data_, TY_Byte error_)
{
    return EN_Status::GETTING_GARBAGE;
}

// Query Functions
EN_Bool CFrame::IsDelimiter(TY_Byte byte_)
{
    TY_Byte u8_Del = byte_ & 0x07;

    switch (u8_Del)
    {
    case CHart::CDelimiter::ACK:
        return EN_Bool::TRUE8;
    case CHart::CDelimiter::STX:
        return EN_Bool::TRUE8;
    case CHart::CDelimiter::BACK:
        return EN_Bool::TRUE8;
    }
    return EN_Bool::FALSE8;
}

EN_Bool CFrame::IsPreamble(TY_Byte byte_)
{
    if (byte_ == 0xff)
    {
        return EN_Bool::TRUE8;
    }
    return EN_Bool::FALSE8;
}

EN_Bool CFrame::IsPrimaryMaster(TY_Byte address_)
{
    if ((address_ & 0x80) != NULL)
    {
        return EN_Bool::TRUE8;
    }
    return EN_Bool::FALSE8;
}

EN_Bool CFrame::IsBurstModeFlag(TY_Byte address_)
{
    if ((address_ & 0x40) != NULL)
    {
        return EN_Bool::TRUE8;
    }
    return EN_Bool::FALSE8;
}

EN_Bool CFrame::IsLongAddress(TY_Byte delimiter_)
{
    if ((delimiter_ & CHart::CMask::ADDR_TYPE) != NULL)
    {
        return EN_Bool::TRUE8;
    }
    return EN_Bool::FALSE8;
}

// HART State Machine Query Functions
EN_Bool CFrame::IsSTX()
{
    return EN_Bool::FALSE8;
}

EN_Bool CFrame::IsACK()
{
    if (GetType() == EN_Type::REQUEST)
    {
        if (m_local_master == m_remote_master)
        {
            return EN_Bool::TRUE8;
        }
    }
    return EN_Bool::FALSE8;
}

EN_Bool CFrame::IsOACK()
{
    if (GetType() == EN_Type::RESPONSE)
    {
        if (m_local_master != m_remote_master)
        {
            return EN_Bool::TRUE8;
        }
    }
    return EN_Bool::FALSE8;
}

EN_Bool CFrame::IsBACK()
{
    if (GetType() == EN_Type::BURST)
    {
        if (m_local_master == m_remote_master)
        {
            return EN_Bool::TRUE8;
        }
    }
    return EN_Bool::FALSE8;
}

EN_Bool CFrame::IsOBACK()
{
    if (GetType() == EN_Type::BURST)
    {
        if (m_local_master != m_remote_master)
        {
            return EN_Bool::TRUE8;
        }
    }
    return EN_Bool::FALSE8;
}

// RxSM Query Functions
TY_Word CFrame::GetRcvByteCount()
{
    return m_rcv_byte_count;
}

TY_Word CFrame::GetTotalLen(TY_Word u16_NewLen)
{
    return (TY_Word)(u16_NewLen + m_rcv_byte_count);
}

// Offline Parsing
void CFrame::ParseData(CFrame* frame_)
{
    // Not yet used
}

// Helpers
void CFrame::ClearBuffers()
{
    m_payload_size = 0;
}

// Setting Frame Data
void CFrame::SetCommand(TY_Byte command_)
{
    m_cmd = command_;
}

void CFrame::SetAddrMode(TY_Byte addr_mode_)
{
    m_addr_mode = addr_mode_;
}

void CFrame::SetNumPreambles(TY_Byte num_preambles_)
{
    m_num_preambles = num_preambles_;
}

void CFrame::SetData(TY_Byte* data_, TY_Word len_)
{
    if (len_ > 0)
    {
        m_payload_size = len_;
        COSAL::CMem::Copy(m_payload_data, data_, len_);
    }
}

void CFrame::SetUniqueID(TY_Byte* bytes_of_unique_id_)
{
    COSAL::CMem::Copy(m_address, bytes_of_unique_id_, 5);
}

void CFrame::SetShortAddr(TY_Byte short_addr_)
{
    m_address[0] = short_addr_;
}

void CFrame::SetShortTag(TY_Byte* short_tag_)
{
    COSAL::CMem::Copy(m_short_tag, short_tag_, 6);
}

void CFrame::SetLongTag(TY_Byte* long_tag_)
{
    COSAL::CMem::Copy(m_long_tag, long_tag_, 24);
}

void CFrame::SetBurstFrame(EN_Bool burst_frame_)
{
    m_burst_frame = burst_frame_;
}

void CFrame::SetLocalMaster(EN_Bool primary_master_)
{
    m_local_master = primary_master_;
}

// Handle time stamps
TY_DWord CFrame::GetStartTime()
{
    return m_start_time;
}

TY_DWord CFrame::GetEndTime()
{
    return m_end_time;
}

void CFrame::SetStartTime(TY_DWord time_)
{
    m_start_time = time_;
}

void CFrame::SetEndTime(TY_DWord time_)
{
    m_end_time = time_;
}

// Getting Frame Data
TY_Byte CFrame::GetCommand()
{
    return m_cmd;
}

TY_Byte CFrame::GetAddrMode()
{
    return m_addr_mode;
}

TY_Byte CFrame::GetNumPreambles()
{
    return m_num_preambles;
}

void CFrame::GetData(TY_Byte* data_, TY_Word* len_)
{
    if ((m_payload_data != 0) && (m_payload_size > 0))
    {
        COSAL::CMem::Copy(data_, m_payload_data, m_payload_size);
        *len_ = m_payload_size;
    }
    else
    {
        *len_ = 0;
    }
}

void CFrame::GetUniqueID(TY_Byte* bytes_of_unique_id_)
{
    COSAL::CMem::Copy(bytes_of_unique_id_, m_address, 5);
}

TY_Byte CFrame::GetShortAddr()
{
    return m_address[0];
}

void CFrame::GetShortTag(TY_Byte* short_tag_)
{
    COSAL::CMem::Copy(short_tag_, m_short_tag, 6);
}

void CFrame::GetLongTag(TY_Byte* long_tag_)
{
    COSAL::CMem::Copy(long_tag_, m_long_tag, 24);
}

EN_Bool CFrame::GetBurstFrame()
{
    return m_burst_frame;
}

EN_Bool CFrame::GetLocalMaster()
{
    return m_local_master;
}

CFrame::EN_Status CFrame::GetStatus()
{
    return m_status;
}

CFrame::EN_Type CFrame::GetType()
{
    switch (m_status)
    {
    case EN_Status::CHECKSUM_ERR:
    case EN_Status::INVALID_FRAME:
        return EN_Type::JUNK;
        break;
    case EN_Status::FRAME_COMPLETED:
        if (m_delimiter == CHart::CDelimiter::STX)
        {
            return EN_Type::REQUEST;
        }
        else if (m_delimiter == CHart::CDelimiter::ACK)
        {
            return EN_Type::RESPONSE;
        }
        else if (m_delimiter == CHart::CDelimiter::BACK)
        {
            return EN_Type::BURST;
        }
        else
        {
            return EN_Type::JUNK;
        }
        break;
    default:
        return EN_Type::UNKNOWN;
        break;
    }
}

TY_Byte* CFrame::GetDataBuffer()
{
    return m_payload_data;
}

TY_Byte CFrame::GetDataSize()
{
    return (TY_Byte)m_payload_size;
}

EN_Bool CFrame::IsBurstModeDevice()
{
    return m_burst_mode;
}

TY_Byte CFrame::GetRspCode1()
{
    return m_rsp1;
}

TY_Byte CFrame::GetRspCode2()
{
    return m_rsp2;
}

TY_Byte CFrame::GetJabOctet()
{
    return m_jab_octet;
}

EN_Bool CFrame::GetJabOctetFlag()
{
    return m_jab_octet_flag;
}

// Operation
void CFrame::SetRequestPrimitive(TY_Byte* data_, TY_Word len_)
{
    if (len_ > 0)
    {
        COSAL::CMem::Copy(m_req_primitive, data_, len_);
        m_req_prim_len = len_;
    }
}

TY_Byte* CFrame::GetTxData(TY_Word* len_)
{
    if (m_req_primitive != NULL)
    {
        if (m_req_prim_len > 0)
        {
            *len_ = m_req_prim_len;
            return m_req_primitive;
        }
    }
    return NULL;
}

// Constructor/Destructor
CFrame::CFrame(EN_Bool local_master_)
{
    m_payload_size = 0;
    m_req_prim_len = 0;
    m_status = EN_Status::GET_PREAMBLE;
    m_start_time = 0;
    m_end_time = 0;
    m_burst_frame = EN_Bool::FALSE8;
    m_burst_mode = EN_Bool::FALSE8;
    m_local_master = local_master_;
    m_rcv_byte_count = 0;
}

CFrame::CFrame()
{
    m_payload_size = 0;
    m_req_prim_len = 0;
    m_status = EN_Status::GET_PREAMBLE;
    m_start_time = 0;
    m_end_time = 0;
    m_burst_frame = EN_Bool::FALSE8;
    m_burst_mode = EN_Bool::FALSE8;
    m_local_master = EN_Bool::FALSE8;
    m_rcv_byte_count = 0;
}

CFrame::~CFrame()
{
}

CFrame& CFrame::operator=(const CFrame& other_)
{
    if (this != &other_)
    {
        if (other_.m_payload_size > 0)
        {
            // Allocate resource and copy to the new data area if other has payload data
            m_payload_size = other_.m_payload_size;
            COSAL::CMem::Copy(m_payload_data, other_.m_payload_data, m_payload_size);
        }
        else
        {
            m_payload_size = 0;
        }

        if (other_.m_req_prim_len > 0)
        {
            // Allocate resource and copy to the new data area if other has payload data
            m_req_prim_len = other_.m_req_prim_len;
            COSAL::CMem::Copy(m_req_primitive, other_.m_req_primitive, m_req_prim_len);
        }
        else
        {
            m_req_prim_len = 0;
        }
        m_status = other_.m_status;
        m_num_preambles = other_.m_num_preambles;
        m_delimiter = other_.m_delimiter;
        m_addr_type = other_.m_addr_type;
        COSAL::CMem::Copy(m_address, other_.m_address, 5);
        m_addr_byte_count = other_.m_addr_byte_count;
        m_num_exp_bytes = other_.m_num_exp_bytes;
        m_exp_bytes_count = other_.m_exp_bytes_count;
        COSAL::CMem::Copy(m_exp_bytes, other_.m_exp_bytes, 3);
        m_local_master = other_.m_local_master;
        m_remote_master = other_.m_remote_master;
        m_cmd = other_.m_cmd;
        m_rsp1 = other_.m_rsp1;
        m_rsp2 = other_.m_rsp2;
        m_payload_count = other_.m_payload_count;
        m_target_chk = other_.m_target_chk;
        m_actual_chk = other_.m_actual_chk;
        m_start_time = other_.m_start_time;
        m_end_time = other_.m_end_time;
        COSAL::CMem::Copy(m_short_tag, other_.m_short_tag, 6);
        COSAL::CMem::Copy(m_long_tag, other_.m_long_tag, 24);
        m_addr_mode = other_.m_addr_mode;
        m_burst_frame = other_.m_burst_frame;
        m_burst_mode = other_.m_burst_mode;
        m_rcv_byte_count = other_.m_rcv_byte_count;
    }

    return *this;
}
