/*
 *          File: HartFrame.h (CFrame)
 *                The hart frame is a construct used to collect all information
 *                which is needed to encode and decode data of so called
 *                service primitives like responses and requests, which are
 *                finally octet streams.
 *
 *        Author: Walter Borst
 *
 *        E-Mail: info@borst-automation.de
 *          Home: https://www.borst-automation.de
 *
 * No Warranties: https://www.borst-automation.com/legal/warranty-disclaimer
 *
 * Copyright 2006-2024 Walter Borst, Cuxhaven, Germany
 */

#ifndef __hartframe_h__
#define __hartframe_h__

#include "OSAL.h"
#include "WbHartUser.h"

class CFrame
{
public:
    // Enum classes
    enum class EN_Status : TY_Byte
    {
        // Parsing status
        GET_PREAMBLE = 0,
        GETTING_PREAMBLES = 1,
        GET_DELIMITER = 2,
        GET_ADDRESS = 3,
        GETTING_LONG_ADDR = 4,
        GET_EXP_BYTES = 5,
        GETTING_EXP_BYTES = 6,
        GET_CMD = 7,
        GET_DATA_LEN = 8,
        GET_RESP1 = 9,
        GET_RESP2 = 10,
        GET_PAYLOAD = 11,
        GETTING_PAYLOAD = 12,
        GET_CHECKSUM = 13,
        GET_GARBAGE = 14,
        GETTING_GARBAGE = 15,
        FRAME_COMPLETED = 16,
        INVALID_FRAME = 17,
        CHECKSUM_ERR = 18
    };
    enum class EN_Type : TY_Byte
    {
        UNKNOWN = 0,
        JUNK = 1,
        REQUEST = 2,
        RESPONSE = 3,
        BURST = 4
    };

    // Initialization/Construction
                          CFrame(EN_Bool local_master_);
                          CFrame();
                         ~CFrame();
               CFrame& operator=(const CFrame& other_);
    void                    Init();
    void                  Uninit();
    // Operation
    EN_Bool             IsActive();
    void               SetActive();
    void             ClearActive();
    EN_Bool             TryParse(TY_Word* bytes_parsed_, TY_Byte* new_data_, TY_Byte* new_err_, TY_Word new_data_len_, EN_Bool gap_time_out_);
    static void        ParseData(CFrame* frame_);
    void     SetRequestPrimitive(TY_Byte* data_, TY_Word len_);
    TY_Byte*           GetTxData(TY_Word* len_);
    EN_Status          GetStatus();
    EN_Type              GetType();
    /* handle time stamps */
    TY_DWord        GetStartTime();
    TY_DWord          GetEndTime();
    void            SetStartTime(TY_DWord time_);
    void              SetEndTime(TY_DWord time_);
    /* Setting Frame Data */
    void              SetCommand(TY_Byte command_);
    void             SetAddrMode(TY_Byte addr_mode_);
    void         SetNumPreambles(TY_Byte num_preambles_);
    void                 SetData(TY_Byte* data_, TY_Word len_);
    void             SetUniqueID(TY_Byte* bytes_of_unique_id_);
    void            SetShortAddr(TY_Byte short_addr_);
    void             SetShortTag(TY_Byte* short_tag_);
    void              SetLongTag(TY_Byte* long_tag_);
    void           SetBurstFrame(EN_Bool burst_frame_);
    void          SetLocalMaster(EN_Bool primary_master_);
    /* Getting Frame Data */
    TY_Byte           GetCommand();
    TY_Byte          GetAddrMode();
    TY_Byte      GetNumPreambles();
    void                 GetData(TY_Byte* data_, TY_Word* len_);
    void             GetUniqueID(TY_Byte* bytes_of_unique_id_);
    TY_Byte         GetShortAddr();
    void             GetShortTag(TY_Byte* short_tag_);
    void              GetLongTag(TY_Byte* long_tag_);
    EN_Bool        GetBurstFrame();
    EN_Bool          GetAnyFrame();
    EN_Bool       GetLocalMaster();
    TY_Byte*       GetDataBuffer();
    TY_Byte          GetDataSize();
    EN_Bool              IsHart6();
    EN_Bool              IsHart7();
    EN_Bool    IsBurstModeDevice();
    TY_Byte          GetRspCode1();
    TY_Byte          GetRspCode2();
    TY_Byte          GetJabOctet();
    EN_Bool      GetJabOctetFlag();
    EN_Bool                IsSTX();
    EN_Bool                IsACK();
    EN_Bool               IsOACK();
    EN_Bool               IsBACK();
    EN_Bool              IsOBACK();
    TY_Word      GetRcvByteCount();
    TY_Word          GetTotalLen(TY_Word new_len_);

private:
    // Parsing status handling
    EN_Status        GetPreamble(TY_Byte data_, TY_Byte error_);
    EN_Status   GettingPreambles(TY_Byte data_, TY_Byte error_);
    EN_Status       GetDelimiter(TY_Byte data_, TY_Byte error_);
    EN_Status         GetAddress(TY_Byte data_, TY_Byte error_);
    EN_Status GettingLongAddress(TY_Byte data_, TY_Byte error_);
    EN_Status        GetExpBytes(TY_Byte data_, TY_Byte error_);
    EN_Status    GettingExpBytes(TY_Byte data_, TY_Byte error_);
    EN_Status             GetCmd(TY_Byte data_, TY_Byte error_);
    EN_Status         GetDataLen(TY_Byte data_, TY_Byte error_);
    EN_Status       GetResponse1(TY_Byte data_, TY_Byte error_);
    EN_Status       GetResponse2(TY_Byte data_, TY_Byte error_);
    EN_Status         GetPayload(TY_Byte data_, TY_Byte error_);
    EN_Status     GettingPayload(TY_Byte data_, TY_Byte error_);
    EN_Status        GetCheckSum(TY_Byte data_, TY_Byte error_);
    EN_Status         GetGarbage(TY_Byte data_, TY_Byte error_);
    EN_Status     GettingGarbage(TY_Byte data_, TY_Byte error_);
    // Query functions
    static EN_Bool     IsDelimiter(TY_Byte byte_);
    static EN_Bool      IsPreamble(TY_Byte byte_);
    static EN_Bool IsPrimaryMaster(TY_Byte address_);
    static EN_Bool IsBurstModeFlag(TY_Byte address_);
    static EN_Bool   IsLongAddress(TY_Byte delimiter_);
    // Helper functions
    void              ClearBuffers();
    // Query functions
    TY_Byte           GetFrameType(CFrame* frame_, TY_Byte byte_);
    TY_Word             GetDataLen(CFrame* frame_, TY_Byte byte_);
    /* Data */
  // Control
    EN_Status m_status;
    // Frame info
    TY_Byte   m_num_preambles;
    TY_Byte   m_delimiter;
    TY_Byte   m_addr_type;
    TY_Byte   m_address[5];
    TY_Byte   m_addr_byte_count;
    TY_Byte   m_num_exp_bytes;
    TY_Byte   m_exp_bytes_count;
    TY_Byte   m_exp_bytes[3];
    EN_Bool   m_local_master;
    EN_Bool   m_remote_master;
    TY_Byte   m_cmd;
    TY_Byte   m_rsp1;
    TY_Byte   m_rsp2;
    TY_Word   m_payload_count;
    TY_Byte   m_payload_data[MAX_PAYLOAD_SIZE];
    TY_Word   m_payload_size;
    TY_Byte   m_target_chk;
    TY_Byte   m_actual_chk;
    TY_DWord  m_start_time;
    TY_DWord  m_end_time;
    TY_Byte   m_short_tag[6];
    TY_Byte   m_long_tag[24];
    TY_Byte   m_addr_mode;
    TY_Byte   m_req_primitive[MAX_TXRX_SIZE];
    TY_Word   m_req_prim_len;
    EN_Bool   m_burst_frame;
    EN_Bool   m_burst_mode;
    TY_Word   m_rcv_byte_count;
    TY_Byte   m_jab_octet;
    EN_Bool   m_jab_octet_flag;
    EN_Bool   m_active;
};

#endif // __hartframe_h__