﻿/*
 *          File: OSAL.cpp
 *                The OSAL concept cannot be applied to all functions that
 *                are required. These functions were implemented in the code
 *                of this module.
 *
 *        Author: Walter Borst
 *
 *        E-Mail: info@borst-automation.de
 *          Home: https://www.borst-automation.de
 *
 * No Warranties: https://www.borst-automation.com/legal/warranty-disclaimer
 *
 * Copyright 2006-2024 Walter Borst, Cuxhaven, Germany
 */
 
#include <Windows.h>
#include "OSAL.h"
#include "WbHartUser.h"
#include "WinSystem.h"
#include "HartM_UartIface.h"
#include <stdio.h>
#include <timeapi.h>
#include <minwindef.h>
#include "HMMacPort.h"

static ST_UartPortData uart_port_data = { NULL, 0, EN_Bool::FALSE8 };
static CWinSys::CThread::ST_ThreadContr uart_thread_control;
void* CWinSys::CUart::m_data;
static CWinSys::CUartTask CTask_uart_cyclic_task;

EN_Bool CWinSys::CUart::Open(TY_Byte com_port_, TY_DWord baudrate_)
{
    TY_Char szPort[15];

    //Open the port at WINAPI
    if (com_port_ > 9)
    {
        sprintf(szPort, "\\\\.\\COM%d", com_port_);
    }
    else
    {
        sprintf(szPort, "COM%d", com_port_);
    }

    uart_port_data.Handle = 
        CreateFile(szPort, GENERIC_READ | GENERIC_WRITE,
            0,                    // exclusive access
            NULL,                 // no security attrs
            OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL,
            NULL
        );
    if (uart_port_data.Handle == INVALID_HANDLE_VALUE)
    {
        // Error stop here
        return EN_Bool::FALSE8;
    }
    else
    {
        COMMTIMEOUTS   strCommTimeOuts;
        DCB            dcb;

        //Initialize the com port interface
        uart_port_data.ByteTime = COSAL::CTimer::GetByteTime(baudrate_);
        //get any early notifications
        SetCommMask(uart_port_data.Handle, EV_RXCHAR);
        //setup device buffers
        SetupComm(uart_port_data.Handle, 256, 256);
        //purge any information in the buffer
        PurgeComm(uart_port_data.Handle,
            PURGE_TXABORT | PURGE_RXABORT |
            PURGE_TXCLEAR | PURGE_RXCLEAR
        );
        //set up for overlapped I/O
        strCommTimeOuts.ReadIntervalTimeout = 20;
        strCommTimeOuts.ReadTotalTimeoutMultiplier = 0;
        strCommTimeOuts.ReadTotalTimeoutConstant = 0;
        strCommTimeOuts.WriteTotalTimeoutMultiplier = 0;
        strCommTimeOuts.WriteTotalTimeoutConstant = 0;
        SetCommTimeouts(uart_port_data.Handle, &strCommTimeOuts);
        //set the com port parameter
        GetCommState(uart_port_data.Handle, &dcb);
        dcb.BaudRate = baudrate_;
        dcb.ByteSize = 8;
        dcb.Parity = ODDPARITY;
        dcb.StopBits = ONESTOPBIT;
        dcb.fBinary = TRUE;
        dcb.fParity = TRUE;
        dcb.fErrorChar = FALSE;
        dcb.ErrorChar = (char)0;
        dcb.fAbortOnError = FALSE;
        dcb.fDtrControl = DTR_CONTROL_ENABLE;
        dcb.fRtsControl = RTS_CONTROL_ENABLE;
        if (SetCommState(uart_port_data.Handle, &dcb) == FALSE)
        {
            Close();
            return EN_Bool::FALSE8;
        }
    }

    // Start the task for the uart handler
    CTask_uart_cyclic_task.Start(CUartMaster::FastCyclicHandler);

    return EN_Bool::TRUE8;
}

void CWinSys::CUart::Close()
{
    if (uart_port_data.Handle != INVALID_HANDLE_VALUE)
    {
        COSAL::Unlock();
        CloseHandle(uart_port_data.Handle);
        uart_port_data.Handle = INVALID_HANDLE_VALUE;
        CWinSys::CThread::Terminate(&uart_thread_control);
        COSAL::Lock();
    }
}

TY_Len CWinSys::CUart::Rx(TY_Word max_len_, ST_RcvByte* rcv_bytes_)
{
    COMSTAT                com_stat;
    int                   read_stat = FALSE;
    DWORD               error_flags = 0;
    DWORD                    length = 0;
    TY_Byte                     err = 0;

    // only try to read number of bytes in queue
    ClearCommError(uart_port_data.Handle, &error_flags, &com_stat);
    if (error_flags)
    {
        if (error_flags & CE_FRAME)
            err |= COSAL::CRcvErr::ERR_Frame;
        if (error_flags & CE_OVERRUN)
            err |= COSAL::CRcvErr::ERR_Over;
        if (error_flags & CE_RXPARITY)
            err |= COSAL::CRcvErr::ERR_Parity;
    }

    length = min(max_len_, com_stat.cbInQue);
    if (length > 0)
    {
        TY_Byte  data[MAX_TXRX_SIZE];
        TY_Word        e;
        TY_DWord time = COSAL::CTimer::GetTime() - (length * uart_port_data.ByteTime);
        DWORD    in_len = length;

        read_stat = ReadFile(uart_port_data.Handle,
            data,
            length,
            &length,
            NULL);
        if (!read_stat)
        {
            // I/O error ???
            return 0;
        }

        // Handle the received data bytes
        for (e = 0; e < length; e++)
        {
            rcv_bytes_[e].Data = data[e];
            rcv_bytes_[e].Error = err;
            rcv_bytes_[e].Time = time;
            time += uart_port_data.ByteTime;
        }
    }

    return (TY_Len)length;
}

EN_Error CWinSys::CUart::Tx(TY_Byte* data_, TY_Word len_)
{
    COMSTAT                com_stat;
    int                   fReadStat = FALSE;
    DWORD               error_flags = 0;
    DWORD                   written = 0;
    DWORD                    length = 0;

    ClearCommError(uart_port_data.Handle, &error_flags, &com_stat);
    WriteFile(uart_port_data.Handle,
        data_,
        len_,
        &written,
        NULL
    );

    return EN_Error::NONE;
}

void CWinSys::CUart::SetCarrierOn()
{
    uart_port_data.CarrOn = EN_Bool::TRUE8;
    EscapeCommFunction(uart_port_data.Handle, SETRTS); //Request to Send
    EscapeCommFunction(uart_port_data.Handle, CLRDTR); //Data Terminal Ready
}

void CWinSys::CUart::SetCarrierOff()
{
    uart_port_data.CarrOn = EN_Bool::FALSE8;
    EscapeCommFunction(uart_port_data.Handle, CLRRTS); //Request to Send
    EscapeCommFunction(uart_port_data.Handle, SETDTR); //Data Terminal Ready
}

EN_Bool CWinSys::CUart::IsCarrierOn(void)
{
    return uart_port_data.CarrOn;
}

EN_Error CWinSys::CThread::Start(CWinSys::CThread::ST_ThreadContr* thread_contr_)
{
    unsigned long ulThreadID;
    void*            hThread;

    thread_contr_->Terminated = EN_Bool::FALSE8;

    hThread = CreateThread((LPSECURITY_ATTRIBUTES)NULL, //No security attributes
        0,                            //Stack size = default
        CWinSys::CThread::Execute,
        thread_contr_,
        0,                            //init flags (0 for running)
        &ulThreadID
    );
    if (hThread != NULL)
    {
        if (thread_contr_->Priority == CPrio::High)
        {
            SetThreadPriority(hThread, THREAD_PRIORITY_TIME_CRITICAL);
        }
        
        // Initialze the central timer
        COSAL::CTimer::Init();

        return EN_Error::NONE;
    }
    else
    {
        return EN_Error::ERR;
    }

    return EN_Error::NONE;
}

void CWinSys::CThread::Terminate(CWinSys::CThread::ST_ThreadContr* task_contr_)
{
    uart_thread_control.RunFlag = EN_Bool::FALSE8;
}

unsigned long __stdcall CWinSys::CThread::Execute(void* data_)
{
    CWinSys::CThread::ST_ThreadContr* thread_control = (ST_ThreadContr*)data_;

    timeBeginPeriod(thread_control->Cycle);
    thread_control->LastTimeMs = timeGetTime();

    while (thread_control->RunFlag == EN_Bool::TRUE8)
    {
        TY_DWord time = timeGetTime();
        TY_DWord passed_time = time - thread_control->LastTimeMs;

        // Correct passed time in debugging sessions
        if (passed_time > (TY_DWord)(10 * thread_control->Cycle))
        {
            thread_control->LastTimeMs = time;
            continue;
        }

        if (passed_time > 0)
        {
            // Set last time to current time and
            // call the handling routine
            thread_control->LastTimeMs = time;

            if (thread_control->Handler != NULL)
            {
                thread_control->Handler(passed_time);
            }
        }
        else
        {
            passed_time = 0;
        }

        if ((passed_time == thread_control->Cycle) || (passed_time == 0))
        {
            Sleep(thread_control->Cycle);
        }
        else if (passed_time < thread_control->Cycle)
        {
            TY_DWord diff = thread_control->Cycle - passed_time;
            Sleep(thread_control->Cycle + diff);
        }
        else
        {
            TY_DWord diff = passed_time - thread_control->Cycle;
            if (thread_control->Cycle > diff)
            {
                Sleep(thread_control->Cycle - diff);
            }
        }

    }

    thread_control->Terminated = EN_Bool::TRUE8;

    timeEndPeriod(thread_control->Cycle);
    return 9136;
}

EN_Error CWinSys::CUartTask::Start(void (*handler_)(TY_Word time_))
{
    // Start the thread for the uart handler
    uart_thread_control.RunFlag = EN_Bool::TRUE8;
    uart_thread_control.Handler = CUartMaster::FastCyclicHandler;
    uart_thread_control.Terminated = EN_Bool::FALSE8;
    uart_thread_control.Priority = 1;
    uart_thread_control.Cycle = 1;

    return CWinSys::CThread::Start(&uart_thread_control);
}

EN_Bool CWinSys::CUartTask::IsTerminated()
{
    if (uart_thread_control.Terminated == EN_Bool::TRUE8)
    {
        m_terminated = EN_Bool::TRUE8;
    }

    return m_terminated;
}

void CUartMaster::FastCyclicHandler(TY_Word time_ms_)
{
#ifdef DEBUG_TIMING
    static TY_DWord  last_time = timeGetTime();
    static TY_DWord   time_diff[101];
    static TY_DWord  time_diff2[101];
    static TY_DWord  start_time = 0;
    static TY_Byte        count = 0;
    static TY_DWord         sum = 0;
    static TY_DWord        sum2 = 0;
    static TY_DWord        time = timeGetTime();
    static TY_DWord        diff = 0;

    time = timeGetTime();

    if (count == 0)
    {
        start_time = time;
        last_time = time;
        sum = 0;
        sum2 = 0;
        count = 1;
        for (int e = 0; e < 101; e++)
        {
            time_diff[e] = 0;
            time_diff2[e] = 0;
        }
    }
    else
    {
        diff = time - last_time;
        time_diff[count - 1] = diff;
        time_diff2[count - 1] = time_ms_;
        sum += diff;
        sum2 += time_ms_;
        count++;
    }

    last_time = time;

    if ((time >= (start_time + 100)) || (count >= 100))
    {
        TY_DWord local_time = COSAL::CTimer::GetTime();
        
        // Set a breakpoint here to check the correct timing
        count = 0;
    }

    /// Additional test in debugging region still required
#endif // DEBUG_TIMING

    // Call the Hart master protocol handler
    CHMMacPort::Execute(time_ms_);
    /// Test
    //if (time_ms_ > 3)
    //{
    //    time_ms_ = 2;
    //}
}
