﻿/*
 *          File: FrmTestMaster.cs (FrmTestMaster)
 *                Includes the operation of the user interface and 
 *                all functions necessary to coordinate the additional 
 *                modules.
 *
 *        Author: Walter Borst
 *
 *        E-Mail: info@borst-automation.de
 *          Home: https://www.borst-automation.de
 *
 * No Warranties: https://www.borst-automation.com/legal/warranty-disclaimer
 *
 * Copyright 2006-2024 Walter Borst, Cuxhaven, Germany
 */

#region Namespaces
using BaTestCppMaster.CommonHelpers;
using BaTestCppMaster.DataSyntax;
using BaTestCppMaster.Forms;
using BaTestCppMaster.HartUtils;
using BaTestCppMaster.Properties;
using System.Collections;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Text;
#endregion

namespace BaTestCppMaster
{
    #region Enumerations
    internal enum EN_Bool : byte
    {
        FALSE8 = 0,
        TRUE8 = 1
    }

    internal enum EN_Endian : byte
    {
        MSB_First = 0,  // Big endian (Hart standard)
        LSB_First = 1   // Little endian
    }
    #endregion Enumerations

    internal partial class FrmTestMaster : Form
    {
        #region Internal Constants
        internal class CLEDcolor
        {
            internal const int BLACK = 0;
            internal const int YELLOW = 1;
            internal const int GREEN = 2;
            internal const int RED = 3;
        }

        internal class CComPortStatus
        {
            internal const int NEUTRAL = 0;
            internal const int FAILED = 1;
            internal const int OK = 2;
            internal const int UNCERTAIN = 3;
        }

        internal class CMonitorStatus
        {
            internal const int NEUTRAL = 0;
            internal const int CONNECTED = 1;
            internal const int UNCERTAIN = 2;
        }

        internal class CLEDstatus
        {
            internal const int IDLE = 0;
            internal const int YELLOW = 1;
            internal const int RED = 2;
            internal const int BLACK = 3;
            internal const int KEEP_RED = 4;
            internal const int FIRST_BLACK = 5;
        }

        internal class CWidthValues
        {
            internal const int COM_LED = 20;
            internal const int MIN_STAT = 100;
            internal const int ERR_LED = 20;
            internal const int NUM = 60;
        }
        #endregion

        #region Private Data
        private int m_number_frames = 0;
        private int m_last_number_frames = -1;
        private bool m_show_menu = true;
        private bool m_last_time_valid = false;
        private uint m_last_end_time = 0;
        private int m_last_color_set = -1;
        private ArrayList mo_hart_frames = new ArrayList();
        private bool m_msg_show_yellow = false;
        private bool m_msg_yellow_busy = false;
        private bool m_msg_show_red = false;
        private bool m_msg_red_busy = false;
        private int m_msg_led_status = CLEDstatus.IDLE;
        private int m_led_cycles = 0;
        // diagnostics
        private int m_num_rcv_frames = 0;
        private int m_num_errors = 0;
        private float m_quality = 100.0f;
        private int m_last_num_rcv_frames = -1;
        private int m_last_num_errors = -1;
        private float m_last_quality = -1.0f;
        // decoding
        private ushort m_data_bytes_selected = 0;
        private string m_selected_data_text = string.Empty;
        private bool m_tool_tip_info_changed = false;
        // status info
        private int m_last_comport_status = -1;
        private int m_last_monitor_status = -1;
        #endregion

        #region ControlLists
        private List<Button> mo_list_of_buttons = new List<Button>();
        private List<CheckBox> mo_list_of_checkboxes = new List<CheckBox>();
        private List<ComboBox> mo_list_of_comboboxes = new List<ComboBox>();
        private List<Button> mo_list_of_hot_buttons = new List<Button>();
        private List<CheckBox> mo_list_of_hot_checkboxes = new List<CheckBox>();

        private void SetupControlLists()
        {
            // Buttons
            mo_list_of_buttons.Clear();
            mo_list_of_buttons.Add(ButSendCmd0short);
            mo_list_of_buttons.Add(butSendCmd1);
            mo_list_of_buttons.Add(ButSendCmd2);
            mo_list_of_buttons.Add(ButSendCmd3);
            mo_list_of_buttons.Add(ButSendCmd12);
            mo_list_of_buttons.Add(ButSendCmd13);
            mo_list_of_buttons.Add(ButSendCmd14);
            mo_list_of_buttons.Add(ButSendCmd15);
            mo_list_of_buttons.Add(ButSendCmd18);
            mo_list_of_buttons.Add(ButSendCmd31);
            mo_list_of_buttons.Add(ButSendCmd35);
            mo_list_of_buttons.Add(ButSendCmd38);
            mo_list_of_buttons.Add(ButSendCmd48);
            mo_list_of_buttons.Add(ButSendUserCmd);
            // Master combo boxes
            mo_list_of_comboboxes.Clear();
            mo_list_of_comboboxes.Add(CmbPollAddress);
            mo_list_of_comboboxes.Add(CmbBaudRate);
            mo_list_of_comboboxes.Add(CmbNumPreambles);
            mo_list_of_comboboxes.Add(CmbMasterRole);
            // Master check boxes
            mo_list_of_checkboxes.Clear();
            // Master hot buttons
            mo_list_of_hot_buttons.Clear();
            mo_list_of_hot_buttons.Add(butSendCmd1);
            mo_list_of_hot_buttons.Add(ButSendCmd2);
            mo_list_of_hot_buttons.Add(ButSendCmd3);
            mo_list_of_hot_buttons.Add(ButSendCmd12);
            mo_list_of_hot_buttons.Add(ButSendCmd13);
            mo_list_of_hot_buttons.Add(ButSendCmd14);
            mo_list_of_hot_buttons.Add(ButSendCmd15);
            mo_list_of_hot_buttons.Add(ButSendCmd18);
            mo_list_of_hot_buttons.Add(ButSendCmd31);
            mo_list_of_hot_buttons.Add(ButSendCmd35);
            mo_list_of_hot_buttons.Add(ButSendCmd38);
            mo_list_of_hot_buttons.Add(ButSendCmd48);
            mo_list_of_hot_buttons.Add(ButSendUserCmd);
            // Master hot check boxes
            mo_list_of_hot_checkboxes.Clear();
        }
        #endregion

        #region Standard Component Management
        internal FrmTestMaster()
        {
            InitializeComponent();
        }
        #endregion

        #region Handle Settings
        private void ApplySettings()
        {
            CMAStatus.ApplyingSettings = true;
            switch (CSettings.ColorSet)
            {
                case 1:
                    RadColorSet2.Checked = true;
                    break;
                case 2:
                    RadColorSetUser.Checked = true;
                    break;
                default:
                    RadColorSet1.Checked = true;
                    break;
            }
            if (CSettings.RetryIfBusy == 1)
            {
                ChkRetryIfBusy.Checked = true;
            }
            else
            {
                ChkRetryIfBusy.Checked = false;
            }
            ChkViewPreambles.Checked = CSettings.ViewPreambles;
            ChkViewAddress.Checked = CSettings.ViewAddressing;
            ChkViewTiming.Checked = CSettings.ViewTiming;
            ChkViewStatusBinary.Checked = CSettings.ViewStatusBinary;
            ChkViewDecodedData.Checked = CSettings.ViewDecodedData;
            ChkViewFrameNumbers.Checked = CSettings.ViewFrameNumbers;
            ChkTopMost.Checked = CSettings.OptionsTopMost;
            CmbNumRetries.SelectedIndex = CSettings.NumRetries;
            if ((CSettings.ComPort > 0) && (CSettings.ComPort < 254))
            {
                CmbComPort.SelectedIndex = CSettings.ComPort - 1;
            }
            else
            {
                CSettings.ComPort = 0;
                CmbComPort.SelectedIndex = 0;
            }

            if ((CSettings.PollAddress >= 0) && (CSettings.PollAddress < 64))
            {
                CmbPollAddress.SelectedIndex = CSettings.PollAddress;
            }
            else
            {
                CSettings.PollAddress = 0;
                CmbPollAddress.SelectedIndex = 0;
            }

            if ((CSettings.BaudRate >= 0) && (CSettings.BaudRate < 64))
            {
                CmbBaudRate.SelectedIndex = CSettings.BaudRate;
            }
            else
            {
                CSettings.BaudRate = 0;
                CmbBaudRate.SelectedIndex = 0;
            }

            if ((CSettings.NumPreambles >= 0) && (CSettings.NumPreambles < 23))
            {
                CmbNumPreambles.SelectedIndex = CSettings.NumPreambles;
            }
            else
            {
                CSettings.NumPreambles = 5;
                CmbNumPreambles.SelectedIndex = 5;
            }

            if ((CSettings.InitialMasterRole >= 0) && (CSettings.InitialMasterRole < 2))
            {
                CmbMasterRole.SelectedIndex = CSettings.InitialMasterRole;
            }
            else
            {
                CSettings.InitialMasterRole = 0;
                CmbMasterRole.SelectedIndex = 0;
            }

            HandleMonitorMode();

            if (this.TopMost != CSettings.OptionsTopMost)
            {
                this.TopMost = CSettings.OptionsTopMost;
            }

            CMAStatus.ApplyingSettings = false;

            // Update all
            CMAStatus.Init();
        }

        private void WriteSettings()
        {
            CSettings.MainFormHeight = this.Height;
            CSettings.MainFormWidth = this.Width;
            CSettings.MainFormLeft = this.Left;
            CSettings.MainFormTop = this.Top;

            // Copy the user colors to default properties
            CSettings.GetColorSetUser();
            // Put internal values to user defaults
            CSettings.Put();
            // Save settings
            Settings.Default.Save();
        }

        private void ReadSettings()
        {
            // Load settings
            Settings.Default.Reload();
            // Read Settings to internal data
            CSettings.Fetch();
            // Adjust main form size and position
            Height = CSettings.MainFormHeight;
            Width = CSettings.MainFormWidth;
            SetMainWindowStartPosition(CSettings.MainFormLeft, CSettings.MainFormTop);
            // Be sure that user colors are not overwritten by standard sets
            CSettings.SaveUserColors();
            // Adjust control values and internal states
            ApplySettings();
        }
        #endregion

        #region Dynamic Configuration and Status Control
        private void CheckConfig()
        {
            CMAStatus.ConfigInProgress = true;
            if (CMAStatus.AnyCfgChanged)
            {
                if (CMAStatus.CMonitor.CfgChanged)
                {
                    CheckForMonitorClose();
                }

                if (CMAStatus.CGeneral.CfgChanged)
                {
                    CheckClientConfig();
                }

                if (CMAStatus.CMonitor.CfgChanged)
                {
                    CheckForMonitorOpen();
                }

                if (CMAStatus.CFramesDisplay.bCfgChanged)
                {
                    DisplayFrames();
                }

                if (CMAStatus.CColors.CfgChanged)
                {
                    ApplyColors();
                    CMAStatus.CfgChanged = true;
                }

                if (CMAStatus.CTopMost.CfgChanged)
                {
                    this.TopMost = CSettings.OptionsTopMost;
                }

                if (CMAStatus.CfgChanged)
                {
                    CheckControlsDisplay();
                }

                SetStatusInfo();
                CMAStatus.ClearChangeFlags();
            }
            CheckMonitorStatus();
            CMAStatus.ConfigInProgress = false;
        }

        private void CheckForMonitorClose()
        {
            if (CMAStatus.CMonitor.CfgChanged)
            {
                HartMasterDLL.BAHAMA_StopMonitor();
                CMAStatus.CGeneral.CfgChanged = true;

                HandleMonitorMode();
            }
        }

        private void CheckForMonitorOpen()
        {
            if (CMAStatus.CMonitor.CfgChanged)
            {
                HartMasterDLL.BAHAMA_StartMonitor();
                HandleMonitorMode();
            }
        }

        private void CheckMonitorConfig()
        {
            if (CMAStatus.CMonitor.CfgChanged)
            {
                if (CMAStatus.CMonitor.ActiveComport != CSettings.ComPort)
                {
                    if (CMAStatus.CMonitor.ActiveComport != 0)
                    {
                        HartMasterDLL.BAHAMA_StopMonitor();
                        HartMasterDLL.BAHAMA_CloseChannel();
                        CMAStatus.CMonitor.ActiveComport = 0;
                    }

                    if (CSettings.ComPort != 0)
                    {
                        HartMasterDLL.BAHAMA_StartMonitor();
                        CMAStatus.CMonitor.ActiveComport = CSettings.ComPort;
                    }
                    CMAStatus.CGeneral.CfgChanged = true;
                }

                HandleMonitorMode();
            }
        }

        private static void CheckClientConfig()
        {
            if (CMAStatus.CGeneral.CfgChanged)
            {
                CTestClient.Configure();
            }
        }

        private void ActivateMonitor()
        {
            if (!CSettings.ModeMonitorFrames)
            {
                CSettings.ModeMonitorFrames = true;
                CMAStatus.CMonitor.CfgChanged = true;
                HandleControlEvent();
            }
        }

        private static bool IsServiceAllowed()
        {
            return CTestClient.IsComPortOpen;
        }

        private static bool IsMonitoringAllowed()
        {
            return true;
        }
        #endregion

        #region Event Procedures
        #region Form
        private void FrmMonitor_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            WriteSettings();
            HartMasterDLL.BAHAMA_StopMonitor();
            CTestClient.Terminate();

            HartMasterDLL.BAHAMA_CloseChannel();
        }

        private void FrmMonitor_Load(object sender, System.EventArgs e)
        {
            // Initialize static classes
            CTestClient.Init(this);
            CFrameHelper.Init();
            //HartDLL.Init();
            CMAStatus.Init();
            SetupControlLists();
            ReadSettings();
            CheckConfig();
            tabMain.SelectedTab = tabMain.TabPages[0];
            DisplayUniqueID();
            SetMsgLedStatusOff();
            rtxtTestDisplay.BorderStyle = BorderStyle.None;
        }

        private void FrmMonitor_Resize(object sender, System.EventArgs e)
        {
            PlaceAndSizeControls();
        }

        private void FrmMonitor_FormClosing(object sender, FormClosingEventArgs e)
        {
            CTestClient.Terminate();
        }
        #endregion Form

        #region TabStart Controls
        private void ChkViewPreambles_CheckedChanged(object sender, EventArgs e)
        {
            ViewPreambles();
        }

        private void ChkViewAddress_CheckedChanged(object sender, EventArgs e)
        {
            ViewAddress();
        }

        private void ChkViewTiming_CheckedChanged(object sender, EventArgs e)
        {
            ViewTiming();
        }

        private void ChkViewFrameNumbers_CheckedChanged(object sender, EventArgs e)
        {
            ViewFrameNumbers();
        }

        private void ChkViewDecodedData_CheckedChanged(object sender, EventArgs e)
        {
            ViewDecodedData();
        }

        private void ChkViewStatusBinary_CheckedChanged(object sender, EventArgs e)
        {
            ViewStatusBinary();
        }

        private void CmbComPort_SelectedIndexChanged(object sender, EventArgs e)
        {
            byte NewComPort;

            if ((CmbComPort.SelectedIndex >= 0) && (CmbComPort.SelectedIndex < 254))
            {
                NewComPort = (byte)(CmbComPort.SelectedIndex + 1);
            }
            else
            {
                NewComPort = 0;
            }
            if (CSettings.ComPort != NewComPort)
            {
                CSettings.ComPort = NewComPort;
                CMAStatus.CMonitor.CfgChanged = true;
                CMAStatus.CGeneral.CfgChanged = true;
                CMAStatus.CfgChanged = true;
                HandleControlEvent();
            }
        }

        private void CmbPollAddress_SelectedIndexChanged(object sender, EventArgs e)
        {
            byte NewPollAddress;

            if ((CmbPollAddress.SelectedIndex >= 0) && (CmbPollAddress.SelectedIndex < 64))
            {
                NewPollAddress = (byte)(CmbPollAddress.SelectedIndex);
            }
            else
            {
                NewPollAddress = 0;
            }
            if (CSettings.PollAddress != NewPollAddress)
            {
                CSettings.PollAddress = NewPollAddress;
                CMAStatus.CGeneral.CfgChanged = true;
                HandleControlEvent();
            }
        }

        private void CmbBaudRate_SelectedIndexChanged(object sender, EventArgs e)
        {
            byte NewBaudRate;

            if ((CmbBaudRate.SelectedIndex >= 0) && (CmbBaudRate.SelectedIndex < 8))
            {
                NewBaudRate = (byte)(CmbBaudRate.SelectedIndex);
            }
            else
            {
                NewBaudRate = 0;
            }
            if (CSettings.BaudRate != NewBaudRate)
            {
                CSettings.BaudRate = NewBaudRate;
                CMAStatus.CGeneral.CfgChanged = true;
                HandleControlEvent();
            }
        }

        private void CmbNumPreambles_SelectedIndexChanged(object sender, EventArgs e)
        {
            byte NewNumPreAmbs;

            if ((CmbNumPreambles.SelectedIndex >= 0) && (CmbNumPreambles.SelectedIndex < 23))
            {
                NewNumPreAmbs = (byte)(CmbNumPreambles.SelectedIndex);
            }
            else
            {
                NewNumPreAmbs = 5;
            }
            if (CSettings.NumPreambles != NewNumPreAmbs)
            {
                CSettings.NumPreambles = NewNumPreAmbs;
                CMAStatus.CGeneral.CfgChanged = true;
                HandleControlEvent();
            }
        }

        private void CmbMasterRole_SelectedIndexChanged(object sender, EventArgs e)
        {
            byte NewMasterRole;

            if ((CmbMasterRole.SelectedIndex >= 0) && (CmbMasterRole.SelectedIndex < 2))
            {
                NewMasterRole = (byte)(CmbMasterRole.SelectedIndex);
            }
            else
            {
                NewMasterRole = 0;
            }
            if (CSettings.InitialMasterRole != NewMasterRole)
            {
                CSettings.InitialMasterRole = NewMasterRole;
                CMAStatus.CGeneral.CfgChanged = true;
                HandleControlEvent();
            }
        }
        #endregion TabStart Controls

        #region TabCommands Controls
        private void ButSendCmd0short_Click(object sender, EventArgs e)
        {
            SendCmd0short();
        }
        private void ButSendCmd1_Click(object sender, EventArgs e)
        {
            SendCmd1();
        }
        private void ButSendCmd2_Click(object sender, EventArgs e)
        {
            SendCmd2();
        }
        private void ButSendCmd3_Click(object sender, EventArgs e)
        {
            SendCmd3();
        }
        private void ButSendCmd12_Click(object sender, EventArgs e)
        {
            SendCmd12();
        }
        private void ButSendCmd13_Click(object sender, EventArgs e)
        {
            SendCmd13();
        }
        private void ButSendCmd14_Click(object sender, EventArgs e)
        {
            SendCmd14();
        }
        private void ButSendCmd15_Click(object sender, EventArgs e)
        {
            SendCmd15();
        }
        private void ButSendCmd18_Click(object sender, EventArgs e)
        {
            SendCmd18();
        }
        private void ButEditCmd18_Click(object sender, EventArgs e)
        {
            Edit_Cmd18();
        }
        private void ButEditCmd31_Click(object sender, EventArgs e)
        {
            Edit_Cmd31();
        }
        private void ButSendCmd31_Click(object sender, EventArgs e)
        {
            SendCmd31();
        }
        private void ButEditUserCmd_Click(object sender, EventArgs e)
        {
            Edit_UserCmd();
        }
        private void ButSendCmd38_Click(object sender, EventArgs e)
        {
            SendCmd38();
        }
        private void ButSendCmd48_Click(object sender, EventArgs e)
        {
            SendCmd48();
        }
        private void ButSendUserCmd_Click(object sender, EventArgs e)
        {
            SendUserCmd();
        }
        #endregion TabCommands Controls

        #region TabOptions Controls
        private void ChkTopMost_CheckedChanged(object sender, EventArgs e)
        {
            MnuFnc_TopMost();
        }

        private void RadColorSet1_CheckedChanged(object sender, EventArgs e)
        {
            CSettings.GetColorSet1();
            CSettings.ColorSet = 0;
            ButEditColors.Enabled = false;
            CMAStatus.CColors.CfgChanged = true;
            HandleControlEvent();
        }

        private void RadColorSet2_CheckedChanged(object sender, EventArgs e)
        {
            CSettings.GetColorSet2();
            CSettings.ColorSet = 1;
            ButEditColors.Enabled = false;
            CMAStatus.CColors.CfgChanged = true;
            HandleControlEvent();
        }

        private void RadColorSetUser_CheckedChanged(object sender, EventArgs e)
        {
            CSettings.GetColorSetUser();
            CSettings.ColorSet = 2;
            ButEditColors.Enabled = true;
            CMAStatus.CColors.CfgChanged = true;
            HandleControlEvent();
        }

        private void CmbNumRetries_SelectedIndexChanged(object sender, EventArgs e)
        {
            byte NewNumRetries = (byte)(CmbNumRetries.SelectedIndex);

            if (CSettings.NumRetries != NewNumRetries)
            {
                CSettings.NumRetries = NewNumRetries;
                CMAStatus.CGeneral.CfgChanged = true;
                HandleControlEvent();
            }
        }

        private void ChkRetryIfBusy_CheckedChanged(object sender, EventArgs e)
        {
            byte NewRetryIfBusy;

            if (ChkRetryIfBusy.Checked == true)
            {
                NewRetryIfBusy = 1;
            }
            else
            {
                NewRetryIfBusy = 0;
            }

            if (CSettings.RetryIfBusy != NewRetryIfBusy)
            {
                CSettings.RetryIfBusy = NewRetryIfBusy;
                CMAStatus.CGeneral.CfgChanged = true;
                HandleControlEvent();
            }
        }

        #endregion TabOptions Controls

        #region Menu Buttons
        private void ButMenu_Click(object sender, EventArgs e)
        {
            if (m_show_menu == true)
            {
                m_show_menu = false;
                butMnuMenu.Text = "Show Tabs";
            }
            else
            {
                m_show_menu = true;
                butMnuMenu.Text = "Hide Tabs";
            }

            PlaceAndSizeControls();
        }

        private void ButMnuInfo_Click(object sender, EventArgs e)
        {
            MnuFnc_InfoAbout();
        }

        private void ButMnuRecord_Click(object sender, EventArgs e)
        {
            MnuFnc_ModeMonitorFrames();
        }

        private void ButMnuClearDisplay_Click(object sender, EventArgs e)
        {
            MnuFnc_ClearDisplay(sender, e);
        }
        #endregion Menu Buttons

        #region Timer
        private void TimWatch_Tick(object sender, EventArgs e)
        {
            HandleWatchingData();
        }

        private void TimOperating_Tick(object sender, EventArgs e)
        {
            HandleWatchingStatus();
        }
        #endregion Timer
        #endregion Event Procedures

        #region Monitor Handling
        internal void HandleMonitorData()
        {
            if (HartMasterDLL.BAHAMA_GetMonitorData(ref HartMasterDLL.MonFrame) != 0)
            {
                if (CSettings.ModeMonitorFrames != false)
                {
                    if (IsMonitoringAllowed())
                    {
                        CFrameData clFrameData = new CFrameData(ref HartMasterDLL.MonFrame);
                        mo_hart_frames.Add(clFrameData);
                        m_number_frames++;
                        UpdateDiagnostics(clFrameData);
                        DisplayFrame(clFrameData, m_number_frames);
                        if ((clFrameData.FrameError) && (clFrameData.MinorError == false))
                        {
                            SetMsgLedColor(CLEDcolor.RED);
                        }
                        else
                        {
                            SetMsgLedColor(CLEDcolor.YELLOW);
                        }
                    }
                }
            }

            CheckMonitorStatus();
        }

        internal void CheckMonitorStatus()
        {
            int monitorStatus = HartMasterDLL.BAHAMA_GetMonitorStatus();

            if (monitorStatus != m_last_monitor_status)
            {
                m_last_monitor_status = monitorStatus;
                switch (monitorStatus)
                {
                    case 0:
                        SetComPortStatus(CComPortStatus.NEUTRAL);
                        break;
                    case 1:
                        SetComPortStatus(CComPortStatus.OK);
                        break;
                    case 2:
                        SetComPortStatus(CComPortStatus.UNCERTAIN);
                        break;
                }
            }
        }
        #endregion Monitor Handling

        #region Service Handling
        internal void UpdateServiceDisplay(int error_)
        {
            string comment = string.Empty;

            switch (error_)
            {
                case CMALib.CConsts.CTClientECodes.NO_ERROR:
                    break;
                case CMALib.CConsts.CTClientECodes.UNKNOWN_ERROR:
                    comment = "Client: Unknown Error Code!";
                    break;
                case CMALib.CConsts.CTClientECodes.COMM_ERROR:
                    comment = "Client: Communication Failure!";
                    break;
                case CMALib.CConsts.CTClientECodes.EMPTY_SERVICE:
                    comment = "Client: Empty Service Used!";
                    break;
                case CMALib.CConsts.CTClientECodes.INVALID_HANDLE:
                    comment = "Client: Invalid Service Handle!";
                    break;
                case CMALib.CConsts.CTClientECodes.NO_DEVICE_RESPONSE:
                    comment = "Client: No Device Response!";
                    SetMsgLedColor(CLEDcolor.RED);
                    break;
                case CMALib.CConsts.CTClientECodes.CONNECTION_FAILED:
                    comment = "Client: Connection Failed!";
                    SetMsgLedColor(CLEDcolor.RED);
                    break;
            }

            if (comment != string.Empty)
            {
                AddCommentToFrame(comment, CSettings.MonColError);
            }
        }
        #endregion

        #region Frame Display Handling
        private void DisplayFrames()
        {
            if (mo_hart_frames != null)
            {
                if (mo_hart_frames.Count > 0)
                {
                    int iFrameNum = 1;

                    rtxtDisplay.Clear();
                    m_last_time_valid = false;
                    foreach (CFrameData clFrmDat in mo_hart_frames)
                    {
                        DisplayFrame(clFrmDat, iFrameNum);
                        iFrameNum++;
                    }
                }
            }
        }

        private void DisplayFrame(CFrameData frm_dat_, int frame_num_)
        {
            int iSelStart = rtxtDisplay.TextLength;

            frm_dat_.FrameNumber = frame_num_;

            if (m_last_time_valid)
            {
                if (m_last_end_time == 0)
                {
                    frm_dat_.RelStartTime = 0xffffffff;
                }
                else
                {
                    if (frm_dat_.StartTime >= m_last_end_time)
                    {
                        frm_dat_.RelStartTime = frm_dat_.StartTime - m_last_end_time;
                    }
                    else
                    {
                        frm_dat_.RelStartTime = 0;
                    }
                }
            }
            else
            {
                frm_dat_.RelStartTime = 0xffffffff;
                m_last_time_valid = true;
            }

            m_last_end_time = frm_dat_.EndTime;

            // Display
            if (frm_dat_.HeadingCommentString != string.Empty)
            {
                string s = string.Empty;
                if (CSettings.ViewTiming)
                {
                    s += "      ";
                }
                if (CSettings.ViewFrameNumbers)
                {
                    s += "     ";
                }
                s += frm_dat_.HeadingCommentString + "\n";
                rtxtDisplay.AppendText(s);
                rtxtDisplay.Select(iSelStart, rtxtDisplay.TextLength - iSelStart);
                rtxtDisplay.SelectionColor = frm_dat_.CommentColor;
                rtxtDisplay.Select(rtxtDisplay.TextLength, 0);
            }

            iSelStart = rtxtDisplay.TextLength;
            rtxtDisplay.AppendText(frm_dat_.GetDisplayString(frame_num_) + "\n");

            #region Colors
            if (CSettings.ViewFrameNumbers)
            {
                // Frame numbers
                rtxtDisplay.Select(iSelStart, 5);
                rtxtDisplay.SelectionColor = CSettings.MonColLinNum;
            }

            if (CSettings.ViewTiming)
            {
                if (CSettings.ViewFrameNumbers)
                {
                    rtxtDisplay.Select(iSelStart + 5, 6);
                }
                else
                {
                    rtxtDisplay.Select(iSelStart, 6);
                }
                rtxtDisplay.SelectionColor = CSettings.MonColTime;
                if (frm_dat_.OffsSelEndTime > 0)
                {
                    rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelEndTime, frm_dat_.LengthEndTime);
                    rtxtDisplay.SelectionColor = CSettings.MonColTime;
                }
            }

            if (frm_dat_.LengthType > 0)
            {
                rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelType, frm_dat_.LengthType);
                if (frm_dat_.PrimaryMaster != false)
                {
                    rtxtDisplay.SelectionColor = CSettings.MonColPrimMaster;
                }
                else
                {
                    rtxtDisplay.SelectionColor = CSettings.MonColScndMaster;
                }
            }

            if (frm_dat_.LengthPreGarbage > 0)
            {
                rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelPreGarbage, frm_dat_.LengthPreGarbage);
                rtxtDisplay.SelectionColor = CSettings.MonColGarbage;
            }

            if (frm_dat_.OffsSelFraming > 0)
            {
                if (frm_dat_.ValidExtCmd)
                {
                    rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelFraming, frm_dat_.LengthFraming + frm_dat_.LengthExtCmd);
                }
                else
                {
                    rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelFraming, frm_dat_.LengthFraming);
                }
                if ((frm_dat_.PreambleValid) && (!frm_dat_.FrameError))
                {
                    rtxtDisplay.SelectionColor = CSettings.MonColHeader;
                    if (frm_dat_.OffsSelChk > 0)
                    {
                        rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelChk, frm_dat_.LengthChk);
                        rtxtDisplay.SelectionColor = CSettings.MonColHeader;
                    }
                }
                else
                {
                    rtxtDisplay.SelectionColor = CSettings.MonColWrongData;
                    if (frm_dat_.OffsSelChk > 0)
                    {
                        rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelChk, frm_dat_.LengthChk);
                        rtxtDisplay.SelectionColor = CSettings.MonColWrongData;
                    }
                }
            }

            if (frm_dat_.ValidExtCmd)
            {
                rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelExtCmd, frm_dat_.LengthExtCmd);
                rtxtDisplay.SelectionColor = CSettings.MonColHeader;
            }

            if (frm_dat_.LengthData > 0)
            {
                rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelData, frm_dat_.LengthData);
                rtxtDisplay.SelectionColor = CSettings.MonColData;
            }

            if (frm_dat_.LengthResp > 0)
            {
                rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelResp, frm_dat_.LengthResp);
                rtxtDisplay.SelectionColor = CSettings.MonColRespIndication;
            }

            if (frm_dat_.LengthError > 0)
            {
                rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelError, frm_dat_.LengthError);
                rtxtDisplay.SelectionColor = CSettings.MonColError;
            }

            if (frm_dat_.LengthGarbage > 0)
            {
                rtxtDisplay.Select(iSelStart + frm_dat_.OffsSelGarbage, frm_dat_.LengthGarbage);
                rtxtDisplay.SelectionColor = CSettings.MonColGarbage;
            }

            rtxtDisplay.Select(rtxtDisplay.TextLength, 0);
            rtxtDisplay.HideSelection = false;
            #endregion

            if (frm_dat_.TailingCommentString != string.Empty)
            {
                if (frm_dat_.TailingCommentString.StartsWith("Client:"))
                {
                    DisplayTailingComment(frm_dat_.TailingCommentString, CSettings.MonColError);
                }
                else
                {
                    DisplayTailingComment(frm_dat_.TailingCommentString, frm_dat_.TailingCommentColor);
                }
            }
        }

        private void DisplayTailingComment(string comment_, Color sel_col_)
        {
            string s = " ";

            if (CSettings.ViewTiming)
            {
                s += "      ";
            }

            if (CSettings.ViewFrameNumbers)
            {
                s += "     ";
            }

            s += comment_ + "\n";
            int iSelStart = rtxtDisplay.TextLength;
            rtxtDisplay.AppendText(s);
            rtxtDisplay.Select(iSelStart, rtxtDisplay.TextLength - iSelStart);
            rtxtDisplay.SelectionColor = sel_col_;
            rtxtDisplay.Select(rtxtDisplay.TextLength, 0);
        }
        #endregion

        #region Internal Methods
        internal bool IsValidComPort(ushort com_port_)
        {
            bool result;

            if (CTestClient.IsComPortOpen)
            {
                if (CSettings.ComPort == com_port_)
                {
                    // No change happened
                    HartMasterDLL.BAHAMA_StopMonitor();
                    result = true;
                }
                else
                {
                    HartMasterDLL.BAHAMA_StopMonitor();
                    HartMasterDLL.BAHAMA_CloseChannel();
                    // Try to open new com port
                    if (HartMasterDLL.BAHAMA_OpenChannel(com_port_) == EN_Bool.TRUE8)
                    {
                        result = true;
                    }
                    else
                    {
                        result = false;
                    }
                }
            }
            else
            {
                // Try to open new com port
                if (HartMasterDLL.BAHAMA_OpenChannel(com_port_) == EN_Bool.TRUE8)
                {
                    result = true;
                }
                else
                {
                    result = false;
                }
            }

            if (result)
            {
                CTestClient.IsComPortOpen = true;
                HartMasterDLL.BAHAMA_StartMonitor();
                SetModeMonitorFrames();
            }
            else
            {
                CTestClient.IsComPortOpen = false;
                ClearModeMonitorFrames();
            }

            return result;
        }
        #endregion

        #region Status Info Handling
        private void SetStatusInfo()
        {
            string s;

            s = "COM " + CSettings.ComPort.ToString();

            if (CSettings.ModeMonitorFrames != false)
            {
                s += " | Monitoring active | " + "Switch record off to stop monitoring.";
                StatusInfo.BackColor = CSettings.StatBarColBkRecording;
            }
            else
            {
                s += " | Monitoring stopped | " + "Switch record on to continue monitoring.";
                StatusInfo.BackColor = CSettings.StatBarColBkNotRec;
            }
            StatusInfo.Items[1].Text = s;
        }
        #endregion

        #region Helpers
        void DisplayTitle()
        {
            Text = CMALib.CGlobText.TEST_CPP_MASTER;
        }

        private void HandleControlEvent()
        {
            if (CMAStatus.HandleEvents)
            {
                CheckConfig();
            }
            if (CSettings.ModeMonitorFrames == false)
            {
                SetMsgLedStatusOff();
            }
        }

        private void AddCommentToFrame(string comment_, Color col_comment_)
        {
            int iLastFrameIdx = mo_hart_frames.Count - 1;

            if (iLastFrameIdx >= 0)
            {
                CFrameData? frame = (CFrameData?)mo_hart_frames[iLastFrameIdx];
                if (frame != null)
                {
                    frame.TailingCommentString = comment_;
                    frame.TailingCommentColor = col_comment_;
                    mo_hart_frames[iLastFrameIdx] = frame;
                    DisplayTailingComment(comment_, col_comment_);
                }
            }
        }
        #endregion

        #region Form Management
        private void SetWindowPosition(int left_, int top_)
        {
            int iDeskTopWidth = SystemInformation.VirtualScreen.Width;
            int iDeskTopHeight = SystemInformation.VirtualScreen.Height;

            if ((left_ + Width) > iDeskTopWidth)
            {
                left_ = iDeskTopWidth - Width;
            }
            if ((top_ + Height) > iDeskTopHeight)
            {
                top_ = iDeskTopHeight - Height;
            }
            Left = left_;
            Top = top_;
        }

        #endregion Form Management

        #region Set/Clear Mode Monitor Frames
        private void SetModeMonitorFrames()
        {
            CSettings.ModeMonitorFrames = true;
            CMAStatus.CMonitor.CfgChanged = true;
            HandleControlEvent();
        }

        private void ClearModeMonitorFrames()
        {
            CSettings.ModeMonitorFrames = false;
            CMAStatus.CMonitor.CfgChanged = true;
            HandleControlEvent();
        }
        #endregion Set/Clear Mode Monitor Frames

        #region Tab Commands Functions
        private void SendCmd0short()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd0short();
            }
        }

        private void SendCmd0short_Wait()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd0short_Wait();
            }
        }

        private void SendCmd1()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd1();
            }
        }
        private void SendCmd2()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd2();
            }
        }

        private void SendCmd3()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd3();
            }
        }
        private void SendCmd3_Wait()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd3_Wait();
            }
        }
        private void SendCmd12()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd12();
            }
        }
        private void SendCmd12_Wait()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd12_Wait();
            }
        }

        private void SendCmd13()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd13();
            }
        }
        private void SendCmd13_Wait()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd13_Wait();
            }
        }
        private void SendCmd14()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd14();
            }
        }

        private void SendCmd14_Wait()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd14_Wait();
            }
        }
        internal void SendCmd15()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd15();
            }
        }

        internal void SendCmd15_Wait()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd15_Wait();
            }
        }

        internal void SendCmd18()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd18(CSettings.TagNameShort, CSettings.Descriptor, CSettings.Day, CSettings.Month, CSettings.Year);
            }
        }

        private void SendCmd20_Wait()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd20_Wait();
            }
        }

        internal void SendCmd35()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd35(CSettings.RangeUnitIdx, CSettings.UpperRange, CSettings.LowerRange);
            }
        }

        private void SendCmd38()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd38();
            }
        }

        private void SendCmd48()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd48();
            }
        }

        internal void SendCmd31()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqCmd31();
            }
        }

        internal void SendUserCmd()
        {
            if (IsServiceAllowed())
            {
                ActivateMonitor();
                CTestClient.ReqUserCmd();
            }
        }
        #endregion Tab Commands Functions

        #region Menu Functions
        #region Start
        private void MnuFnc_ClearDisplay(object sender, System.EventArgs e)
        {
            ClearDisplay();
        }

        private void ClearDisplay()
        {
            rtxtDisplay.Clear();
            mo_hart_frames.Clear();
            // CheckMenuVisibility();
            m_last_end_time = 0;
            m_last_time_valid = false;
            m_number_frames = 0;
            SetStatusInfo();
            DisplayTitle();
            m_last_number_frames = -1;
            ClearDiagnostics();
            rtxtTestDisplay.Clear();
        }

        #endregion Start

        #region View
        private void ViewPreambles()
        {
            if (ChkViewPreambles.Checked != false)
            {
                CSettings.ViewPreambles = true;
            }
            else
            {
                CSettings.ViewPreambles = false;
            }

            CMAStatus.CFramesDisplay.bCfgChanged = true;
            HandleControlEvent();
        }

        private void ViewAddress()
        {
            if (ChkViewAddress.Checked != false)
            {
                CSettings.ViewAddressing = true;
            }
            else
            {
                CSettings.ViewAddressing = false;
            }

            CMAStatus.CFramesDisplay.bCfgChanged = true;
            HandleControlEvent();
        }

        private void ViewTiming()
        {
            if (ChkViewTiming.Checked != false)
            {
                CSettings.ViewTiming = true;
            }
            else
            {
                CSettings.ViewTiming = false;
            }

            CMAStatus.CFramesDisplay.bCfgChanged = true;
            HandleControlEvent();
        }

        private void ViewDecodedData()
        {
            if (ChkViewDecodedData.Checked != false)
            {
                CSettings.ViewDecodedData = true;
            }
            else
            {
                CSettings.ViewDecodedData = false;
            }

            CMAStatus.CFramesDisplay.bCfgChanged = true;
            HandleControlEvent();
        }

        private void ViewFrameNumbers()
        {
            if (ChkViewFrameNumbers.Checked != false)
            {
                CSettings.ViewFrameNumbers = true;
            }
            else
            {
                CSettings.ViewFrameNumbers = false;
            }

            CMAStatus.CFramesDisplay.bCfgChanged = true;
            HandleControlEvent();
        }

        private void ViewStatusBinary()
        {
            if (ChkViewStatusBinary.Checked != false)
            {
                CSettings.ViewStatusBinary = true;
            }
            else
            {
                CSettings.ViewStatusBinary = false;
            }
            CMAStatus.CFramesDisplay.bCfgChanged = true;
            HandleControlEvent();
        }
        #endregion

        #region Options
        private void MnuFnc_TopMost()
        {
            if (ChkTopMost.Checked != false)
            {
                CSettings.OptionsTopMost = true;
            }
            else
            {
                CSettings.OptionsTopMost = false;
            }
            CMAStatus.CTopMost.CfgChanged = true;
            HandleControlEvent();
        }
        #endregion Options

        #region Edit Commands Data
        private void Edit_Cmd18()
        {
            FrmCmd18 f = new FrmCmd18(this);

            f.ShowDialog(this);
        }

        private void Edit_Cmd31()
        {
            FrmCmd31 f = new FrmCmd31(this);

            f.ShowDialog(this);
        }

        private void Edit_Cmd35()
        {
            FrmCmd35 f = new FrmCmd35(this);

            f.ShowDialog(this);
        }

        private void Edit_UserCmd()
        {
            FrmUserCmd f = new FrmUserCmd(this);

            f.ShowDialog(this);
        }
        #endregion

        #region Monitor Mode
        private void MnuFnc_ModeMonitorFrames()
        {
            if (CSettings.ModeMonitorFrames == true)
            {
                CSettings.ModeMonitorFrames = false;
            }
            else
            {
                CSettings.ModeMonitorFrames = true;
            }

            CMAStatus.CMonitor.CfgChanged = true;
            HandleControlEvent();
        }
        #endregion Monitor Mode

        #region Help
        private void MnuFnc_InfoAbout()
        {
            FrmAbout f = new FrmAbout();

            f.lblVersion.Text = "7.6.1";
            f.lblDate.Text = "30.6.2024";
            f.ForceLeft = this.Right - f.Width - 5;
            f.ForceTop = this.Top + 70;
            f.ShowDialog();
        }
        #endregion Help
        #endregion Menu Functions

        #region Appearance Control
        private void PlaceAndSizeControls()
        {
            int leftPos;
            int iTotalWidth = StatusInfo.ClientRectangle.Width;
            int i;

            Rectangle r = ClientRectangle;
            panTop.Top = 0;
            panTop.Left = 0;
            panTop.Width = r.Width;
            butMnuMenu.Left = r.Width - butMnuMenu.Width;
            if (m_show_menu == true)
            {
                tabMain.Top = panTop.Height;
                tabMain.Left = 0;
                tabMain.Width = r.Width;
                ArrangeWorkArea(panTop.Height + tabMain.Height, 0, r.Width, r.Height - panTop.Height - tabMain.Height - StatusInfo.Height);
                tabMain.Visible = true;
            }
            else
            {
                tabMain.Visible = false;
                ArrangeWorkArea(panTop.Height, 0, r.Width, r.Height - panTop.Height - StatusInfo.Height);
            }
            // Sort out the values on the top panel
            int requiredWidth = butMnuClearDisplay.Width + 5 + butMnuRecord.Width + 5 + butMnuInfo.Width + 5 + butMnuMenu.Width;

            if (requiredWidth < r.Width)
            {
                leftPos = r.Width - requiredWidth;
            }
            else
            {
                leftPos = 0;
            }
            butMnuClearDisplay.Left = leftPos;
            leftPos += butMnuClearDisplay.Width + 5;
            butMnuRecord.Left = leftPos;
            leftPos += butMnuRecord.Width + 5;
            butMnuMenu.Left = leftPos;
            leftPos += butMnuMenu.Width + 5;
            butMnuInfo.Left = leftPos;
            i = CWidthValues.COM_LED + CWidthValues.ERR_LED + CWidthValues.NUM;
            if ((i + 100) <= iTotalWidth)
            {
                StatusInfo.Items[1].Width = iTotalWidth - i - 50;
            }
            else
            {
                StatusInfo.Items[1].Width = 100;
            }
            StatusInfo.Items[0].Width = CWidthValues.COM_LED;
            StatusInfo.Items[2].Width = CWidthValues.ERR_LED;
            StatusInfo.Items[3].Width = CWidthValues.NUM;
        }

        private void ArrangeWorkArea(int top_, int left_, int width_, int height_)
        {
            rtxtDisplay.Parent = this;
            rtxtDisplay.Top = top_;
            rtxtDisplay.Left = left_;
            rtxtDisplay.Height = height_;
            rtxtDisplay.Width = width_;
        }
        #endregion Appearance Control

        #region Menu Visibility and Monitor Status Handling
        internal static bool IsSendPossible()
        {
            return CTestClient.IsComPortOpen;
        }

        private void DisableControls()
        {
            foreach (Button b in mo_list_of_buttons)
            {
                b.Enabled = false;
            }
            foreach (CheckBox c in mo_list_of_checkboxes)
            {
                c.Enabled = false;
            }
            foreach (ComboBox c in mo_list_of_comboboxes)
            {
                c.Enabled = false;
            }
            foreach (CheckBox c in mo_list_of_hot_checkboxes)
            {
                c.Enabled = false;
            }
            this.Refresh();
        }

        private void CheckControlsDisplay()
        {
            if (CTestClient.IsComPortOpen)
            {
                foreach (Button b in mo_list_of_buttons)
                {
                    b.Enabled = true;
                }
                foreach (CheckBox c in mo_list_of_checkboxes)
                {
                    c.Enabled = true;
                }
                foreach (ComboBox c in mo_list_of_comboboxes)
                {
                    c.Enabled = true;
                }

                if (CTestClient.IsConnected)
                {
                    ButConnect.BackColor = SystemColors.Control;
                    ButConnect.Enabled = false;
                    ButDisconnect.BackColor = CSettings.ButColBkHot;
                    ButDisconnect.Enabled = true;
                    foreach (Button b in mo_list_of_hot_buttons)
                    {
                        b.BackColor = CSettings.ButColBkHot;
                        b.Enabled = true;
                    }
                    foreach (CheckBox c in mo_list_of_hot_checkboxes)
                    {
                        c.Enabled = true;
                    }
                }
                else
                {
                    ButConnect.BackColor = CSettings.ButColBkHot;
                    ButConnect.Enabled = true;
                    ButDisconnect.BackColor = SystemColors.Control;
                    ButDisconnect.Enabled = false;
                    foreach (Button b in mo_list_of_hot_buttons)
                    {
                        b.BackColor = SystemColors.Control;
                        b.Enabled = false;
                    }
                    foreach (CheckBox c in mo_list_of_hot_checkboxes)
                    {
                        c.Enabled = false;
                    }
                }

                ButSendCmd0short.BackColor = CSettings.ButColBkHot;
                ButSendCmd0short.Enabled = true;
            }
            else
            {
                foreach (Button b in mo_list_of_buttons)
                {
                    b.Enabled = false;
                }
                foreach (CheckBox c in mo_list_of_checkboxes)
                {
                    c.Enabled = false;
                }
                foreach (ComboBox c in mo_list_of_comboboxes)
                {
                    c.Enabled = false;
                }
                foreach (Button b in mo_list_of_hot_buttons)
                {
                    b.BackColor = SystemColors.Control;
                    b.Enabled = false;
                }
                foreach (CheckBox c in mo_list_of_hot_checkboxes)
                {
                    c.Enabled = false;
                }

                ButConnect.BackColor = SystemColors.Control;
                ButConnect.Enabled = false;
                ButDisconnect.BackColor = SystemColors.Control;
                ButDisconnect.Enabled = false;
                ButSendCmd0short.BackColor = SystemColors.Control;
                ButSendCmd0short.Enabled = false;
            }

            if (CTestClient.IsConnected)
            {
                TxtConStatus.Text = "Connected";
            }
            else
            {
                TxtConStatus.Text = "Not Connected";
            }
        }

        #endregion

        #region Handle Monitor Mode
        private void HandleMonitorMode()
        {
            if (CSettings.ModeMonitorFrames)
            {
                butMnuRecord.Text = "Record Off";
            }
            else
            {
                butMnuRecord.Text = "Record On";
            }
        }
        #endregion Handle Monitor Mode

        #region Color Handling
        private void ButEditColors_Click(object sender, EventArgs e)
        {
            FrmSetColors f = new FrmSetColors(this);
            bool bWasTopMost = CSettings.OptionsTopMost;

            if (bWasTopMost)
            {
                this.TopMost = false;
            }

            if (f.ShowDialog() == DialogResult.OK)
            {
                CSettings.SaveUserColors();
                ApplyColorsFromForm(f);
            }
            if (bWasTopMost)
            {
                this.TopMost = true;
            }
        }

        internal void ApplyColorsFromForm(FrmSetColors f_)
        {
            if (f_.MonitorColorsCanged)
            {
                ApplyMonitorColors();
            }

            if (f_.StatusBarColorsCanged)
            {
                ApplyStatusInfoColors();
            }

            if (f_.HotButtonsColorsCanged)
            {
                ApplyHotKeyColors();
            }
        }

        private void ApplyColors()
        {
            if (m_last_color_set != CSettings.ColorSet)
            {
                m_last_color_set = CSettings.ColorSet;
                ApplyMonitorColors();
                ApplyHotKeyColors();
            }
        }

        private void ApplyMonitorColors()
        {
            rtxtDisplay.BackColor = CSettings.MonColBack;
            DisplayFrames();
        }

        private void ApplyStatusInfoColors()
        {
            SetStatusInfo();
        }

        private void ApplyHotKeyColors()
        {
            ButSendCmd0short.BackColor = CSettings.ButColBkHot;
            butSendCmd1.BackColor = CSettings.ButColBkHot;
            ButSendCmd3.BackColor = CSettings.ButColBkHot;
        }
        #endregion Color Handling

        #region Update Data
        internal void UpdatePollAddress(byte poll_address_)
        {
            if (poll_address_ < 64)
            {
                CSettings.PollAddress = poll_address_;
                CmbPollAddress.SelectedIndex = poll_address_;
            }
        }

        internal void UpdateUniqueID(byte[] bytes_of_unique_id_)
        {
            CTestClient.SetUniqueID(bytes_of_unique_id_);
            DisplayUniqueID();
        }
        #endregion

        #region Display Items
        private void DisplayUniqueID()
        {
            byte[] abyTmp = CTestClient.GetUniqueID();

            TxtDevDatUniqueID.Text = "0x" + CFrameHelper.GetHex(abyTmp[0]);
            TxtDevDatUniqueID.Text += CFrameHelper.GetHex(abyTmp[1]);
            TxtDevDatUniqueID.Text += CFrameHelper.GetHex(abyTmp[2]);
            TxtDevDatUniqueID.Text += CFrameHelper.GetHex(abyTmp[3]);
            TxtDevDatUniqueID.Text += CFrameHelper.GetHex(abyTmp[4]);
        }
        #endregion

        #region Handle Indication LEDs
        internal void SetComPortStatus(int status_)
        {
            if (status_ == m_last_comport_status)
            {
                return;
            }

            m_last_comport_status = status_;
            StatusInfo.Items[0].Image = status_ switch
            {
                CComPortStatus.FAILED => ImgL_Leds.Images[CLEDcolor.RED],
                CComPortStatus.OK => ImgL_Leds.Images[CLEDcolor.GREEN],
                CComPortStatus.UNCERTAIN => ImgL_Leds.Images[CLEDcolor.YELLOW],
                _ => ImgL_Leds.Images[CLEDcolor.BLACK],
            };
        }

        internal void HandleMsgLedStatus()
        {
            // Take over request if not busy
            if ((!m_msg_yellow_busy) && (!m_msg_red_busy))
            {
                // YELLOW has priority
                if (!m_msg_red_busy)
                {
                    if (m_msg_show_red)
                    {
                        m_msg_red_busy = true;
                        m_msg_show_red = false;
                        m_led_cycles = 0;
                    }
                }
            }

            if ((!m_msg_yellow_busy) && (!m_msg_red_busy))
            {
                // YELLOW has priority
                if (!m_msg_yellow_busy)
                {
                    if (m_msg_show_yellow)
                    {
                        m_msg_yellow_busy = true;
                        m_msg_show_yellow = false;
                        m_led_cycles = 0;
                    }
                }
            }

            if (m_msg_yellow_busy)
            {
                switch (m_msg_led_status)
                {
                    case CLEDstatus.KEEP_RED:
                        m_led_cycles = 0;
                        // First show black
                        StatusInfo.Items[2].Image = ImgL_Leds.Images[CLEDcolor.BLACK];
                        m_msg_led_status = CLEDstatus.FIRST_BLACK;
                        break;
                    case CLEDstatus.FIRST_BLACK:
                    case CLEDstatus.IDLE:
                        if (m_led_cycles == 7)
                        {
                            m_led_cycles = 0;
                            // Show yellow image
                            StatusInfo.Items[2].Image = ImgL_Leds.Images[CLEDcolor.YELLOW];
                            m_msg_led_status = CLEDstatus.YELLOW;
                        }
                        break;
                    case CLEDstatus.YELLOW:
                        if (m_led_cycles == 1)
                        {
                            m_led_cycles = 0;
                            // Show black image
                            StatusInfo.Items[2].Image = ImgL_Leds.Images[CLEDcolor.BLACK];
                            m_msg_led_status = CLEDstatus.BLACK;
                        }
                        break;
                    case CLEDstatus.BLACK:
                        if (m_led_cycles == 7)
                        {
                            m_led_cycles = 0;
                            // Done
                            m_msg_led_status = CLEDstatus.IDLE;
                            m_msg_yellow_busy = false;
                        }
                        break;
                }
            }

            if (m_msg_red_busy)
            {
                switch (m_msg_led_status)
                {
                    case CLEDstatus.KEEP_RED:
                        m_led_cycles = 0;
                        // First show black
                        StatusInfo.Items[2].Image = ImgL_Leds.Images[CLEDcolor.BLACK];
                        m_msg_led_status = CLEDstatus.FIRST_BLACK;
                        break;
                    case CLEDstatus.FIRST_BLACK:
                    case CLEDstatus.IDLE:
                        if (m_led_cycles == 3)
                        {
                            m_led_cycles = 0;
                            // Show red image
                            StatusInfo.Items[2].Image = ImgL_Leds.Images[CLEDcolor.RED];
                            m_msg_led_status = CLEDstatus.RED;
                        }
                        break;
                    case CLEDstatus.RED:
                        if (m_led_cycles == 10)
                        {
                            m_led_cycles = 0;
                            // Show black image
                            m_msg_led_status = CLEDstatus.KEEP_RED;
                            m_msg_red_busy = false;
                        }
                        break;
                }
            }

            if (m_led_cycles < 10)
            {
                m_led_cycles++;
            }
        }

        internal void SetMsgLedColor(int status_)
        {
            if (status_ == CLEDcolor.RED)
            {
                m_msg_show_red = true;
            }
            if (status_ == CLEDcolor.YELLOW)
            {
                m_msg_show_yellow = true;
            }
        }

        internal void SetMsgLedStatusOff()
        {
            m_msg_show_yellow = false;
            m_msg_yellow_busy = false;
            m_msg_show_red = false;
            m_msg_red_busy = false;
            m_msg_led_status = CLEDstatus.IDLE;
            m_led_cycles = 0;
            StatusInfo.Items[2].Image = ImgL_Leds.Images[CLEDcolor.BLACK];
        }
        #endregion

        #region Handle Diagnostics
        private void UpdateDiagnostics(CFrameData frame_data_)
        {
            float fNumRcvs;
            float fNumErr;

            if (!frame_data_.ClientTx)
            {
                if (!frame_data_.MinorError)
                {
                    m_num_rcv_frames++;
                    if (frame_data_.FrameError)
                    {
                        m_num_errors++;
                    }
                    fNumRcvs = Convert.ToSingle(m_num_rcv_frames);
                    fNumErr = Convert.ToSingle(m_num_errors);
                    // calculate the quality
                    m_quality = 100.0f - ((fNumErr / fNumRcvs) * 100.0f);
                }
            }
        }

        private void ClearDiagnostics()
        {
            m_num_rcv_frames = 0;
            m_num_errors = 0;
            m_quality = 100.0f;
            m_last_num_rcv_frames = -1;
            m_last_num_errors = -1;
            m_last_quality = -1.0f;
        }

        private void DisplayDiagnostics()
        {
            if (m_last_num_rcv_frames != m_num_rcv_frames)
            {
                m_last_num_rcv_frames = m_num_rcv_frames;
            }
            if (m_last_num_errors != m_num_errors)
            {
                m_last_num_errors = m_num_errors;
            }
            if (m_last_quality != m_quality)
            {
                m_last_quality = m_quality;
            }
        }
        #endregion

        #region Data Decoding
        private void TsmnuDecInteger_Click(object sender, EventArgs e)
        {
            string s;

            if (!CheckForDecoding())
            {
                return;
            }
            if (m_data_bytes_selected > 0)
            {
                s = CFrameHelper.CDecode.DecodeInteger(m_data_bytes_selected);
                ShowToolTip("Integer Value(s)", s);
            }
            else
            {
                CFrameHelper.SignalInvalidSelection(m_selected_data_text);
            }
        }

        private bool CheckForDecoding()
        {
            m_selected_data_text = rtxtDisplay.SelectedText;

            if (m_selected_data_text.Length > 0)
            {
                m_data_bytes_selected = CFrameHelper.GetBytesFromFrameData(m_selected_data_text);

                if (m_data_bytes_selected > 0)
                {
                    return true;
                }
            }
            else
            {
                CFrameHelper.SignalNoSelection();
            }
            return false;
        }

        private void TsmnuDecFloat_Click(object sender, EventArgs e)
        {
            string s;

            if (!CheckForDecoding())
            {
                return;
            }
            if (m_data_bytes_selected > 3)
            {
                s = CFrameHelper.CDecode.DecodeFloat(m_data_bytes_selected);
                ShowToolTip("Float Value(s)", s);
            }
            else
            {
                CFrameHelper.SignalInvalidSelection(m_selected_data_text);
            }
        }

        private void TsmnuDecDouble_Click(object sender, EventArgs e)
        {
            string s;

            if (!CheckForDecoding())
            {
                return;
            }
            if (m_data_bytes_selected > 7)
            {
                s = CFrameHelper.CDecode.DecodeDouble(m_data_bytes_selected);
                ShowToolTip("Double Value(s)", s);
            }
            else
            {
                CFrameHelper.SignalInvalidSelection(m_selected_data_text);
            }
        }

        private void TsmnuDecHartUnit_Click(object sender, EventArgs e)
        {
            string s;

            if (!CheckForDecoding())
            {
                return;
            }
            if (m_data_bytes_selected > 0)
            {
                s = CFrameHelper.CDecode.DecodeHartUnit(m_data_bytes_selected);
                ShowToolTip("Hart Unit(s)", s);
            }
            else
            {
                CFrameHelper.SignalInvalidSelection(m_selected_data_text);
            }
        }

        private void TsmnuDecPackedASCII_Click(object sender, EventArgs e)
        {
            string s;

            if (!CheckForDecoding())
            {
                return;
            }
            if (m_data_bytes_selected > 0)
            {
                s = CFrameHelper.CDecode.DecodePackedASCII(m_data_bytes_selected);
                ShowToolTip("Packed ASCII", s);
            }
            else
            {
                CFrameHelper.SignalInvalidSelection(m_selected_data_text);
            }
        }

        private void TsmnuDecText_Click(object sender, EventArgs e)
        {
            string s;

            if (!CheckForDecoding())
            {
                return;
            }
            if (m_data_bytes_selected > 0)
            {
                s = CFrameHelper.CDecode.DecodeText(m_data_bytes_selected);
                ShowToolTip("Text", s);
            }
            else
            {
                CFrameHelper.SignalInvalidSelection(m_selected_data_text);
            }
        }

        private void TsmnuDecBinary_Click(object sender, EventArgs e)
        {
            string s;

            if (!CheckForDecoding())
            {
                return;
            }
            if (m_data_bytes_selected > 0)
            {
                s = CFrameHelper.CDecode.DecodeBinary(m_data_bytes_selected);
                ShowToolTip("Binary", s);
            }
            else
            {
                CFrameHelper.SignalInvalidSelection(m_selected_data_text);
            }
        }

        private void ShowToolTip(string sTitle, string sCaption)
        {
            ToolTipDecodedData.ToolTipTitle = sTitle;
            //toolTipDecodedData.SetToolTip(rtxtDisplay, sCaption); 
            ToolTipDecodedData.Show(sCaption, rtxtDisplay, 60000);
        }

        private void CtxMnuDecode_Opened(object sender, EventArgs e)
        {
            ToolTipDecodedData.Hide(rtxtDisplay);
        }
        #endregion

        #region Copy to Any Frame
        private void TsmnuBytesToClipBoard_Click(object sender, EventArgs e)
        {
            string s;

            if (!CheckForDecoding())
            {
                return;
            }
            if (m_data_bytes_selected > 0)
            {
                s = CFrameHelper.HexDumpByteDataToDataStx((byte)m_data_bytes_selected);
                Clipboard.SetText(s);
            }
            else
            {
                CFrameHelper.SignalInvalidSelection(m_selected_data_text);
            }
        }
        #endregion

        #region Showing tool tips
        private void ToolTipInfo_Popup(object sender, PopupEventArgs e)
        {
            if (m_tool_tip_info_changed == true)
            {
                return;
            }
            m_tool_tip_info_changed = true;
            if (e.AssociatedControl == ButSendCmd31)
            {
                ToolTipInfo.SetToolTip(ButSendCmd31, "Cmd " + CSettings.ExtCmd.ToString("0") + " | " + CSettings.ExtCmdData);
            }

            m_tool_tip_info_changed = false;
        }
        #endregion Showing tool tips

        #region Main Window Start Position
        private void SetMainWindowStartPosition(int left_, int top_)
        {
            Point initTopLeft = new Point(left_, top_);
            Screen[] screens = Screen.AllScreens;

            foreach (Screen screen in screens)
            {
                if (screen.WorkingArea.Contains(initTopLeft))
                {
                    SetWindowPosition(CSettings.MainFormLeft, CSettings.MainFormTop);
                    return;
                }
            }
            // Window is off screen, center on main desktop
            Screen? desktop = Screen.PrimaryScreen;
            if (desktop != null)
            {
                Rectangle workRect = desktop.WorkingArea;
                int newleft = (workRect.Width - this.Width) / 2;
                int newtop = (workRect.Height - this.Height) / 2;
                SetWindowPosition(newleft, newtop);
            }
        }
        #endregion

        #region Slave Device Data Handling
        private void ButConnect_Click(object sender, EventArgs e)
        {
            ButSendCmd0short_Click(sender, e);
        }

        private void ButDisconnect_Click(object sender, EventArgs e)
        {
            CTestClient.Disconnect();
            CMAStatus.CMonitor.CfgChanged = true;
            CMAStatus.CGeneral.CfgChanged = true;
            CMAStatus.CfgChanged = true;
            HandleControlEvent();
        }

        private void ButSendCmd35_Click(object sender, EventArgs e)
        {
            SendCmd35();
        }

        private void ButEditCmd35_Click(object sender, EventArgs e)
        {
            Edit_Cmd35();
        }

        private void ButDetDatReadAll_Click(object sender, EventArgs e)
        {
            TimOperating.Enabled = false;
            TimWatch.Enabled = false;

            TxtDevDatUniqueID.Text = "-/-";
            TxtDevDatShortTag.Text = "-/-";
            TxtDevDatLongTag.Text = "-/-";
            TxtDevDatMessage.Text = "-/-";
            TxtDevDatUpperRange.Text = "-/-";
            LblDevDatRangeUnitUp.Text = "-/-";
            TxtDevDatLowerRange.Text = "-/-";
            LblDevDatRangeUnitLo.Text = "-/-";
            TxtDevDatSpan.Text = "-/-";
            LblDevDatSpanUnit.Text = "-/-";
            TxtDevDatPV.Text = "-/-";
            LblDevDatUnitPV.Text = "-/-";
            TxtDevDatSV.Text = "-/-";
            LblDevDatUnitSV.Text = "-/-";
            TxtDevDatTV.Text = "-/-";
            LblDevDatUnitTV.Text = "-/-";
            TxtDevDatQV.Text = "-/-";
            LblDevDatUnitQV.Text = "-/-";
            TxtDevDatCurrent.Text = "-/-";
            this.Refresh();

            CTestClient.ReqCmd0short();
            while (HartMasterDLL.BAHAMA_IsServiceCompleted(CTestClient.hServiceConnect) != EN_Bool.TRUE8)
            {
                HandleCyclics();
                Thread.Sleep(50);
            }

            HandleCyclics();

            if (CTestClient.IsConnected)
            {
                SendCmd3_Wait();
                HandleCyclics();
            }

            if (CTestClient.IsConnected)
            {
                SendCmd12_Wait();
                HandleCyclics();
            }

            if (CTestClient.IsConnected)
            {
                SendCmd13_Wait();
                HandleCyclics();
            }

            if (CTestClient.IsConnected)
            {
                SendCmd14_Wait();
                HandleCyclics();
            }

            if (CTestClient.IsConnected)
            {
                SendCmd15_Wait();
                HandleCyclics();
            }

            if (CTestClient.IsConnected)
            {
                SendCmd20_Wait();
                HandleCyclics();
            }

            TimOperating.Enabled = true;
            TimWatch.Enabled = true;
        }
        internal void DetDatShowShortTag(string tag_)
        {
            TxtDevDatShortTag.Text = tag_;
            TxtDevDatShortTag.Refresh();
        }
        internal void DevDatShowLongTag(string tag_)
        {
            TxtDevDatLongTag.Text = tag_;
            TxtDevDatLongTag.Refresh();
        }
        internal void DevDatShowMessage(string message_)
        {
            TxtDevDatMessage.Text = message_;
            TxtDevDatMessage.Refresh();
        }
        internal void DevDatShowUpperRange(float value_, byte unit_)
        {
            TxtDevDatUpperRange.Text = CHelpers.CFormat.Float5digit(value_);
            LblDevDatRangeUnitUp.Text = CHartUtils.CHartUnit.GetText(unit_);
            TxtDevDatUpperRange.Refresh();
            LblDevDatRangeUnitUp.Refresh();
        }
        internal void DevDatShowLowerRange(float value_, byte unit_)
        {
            TxtDevDatLowerRange.Text = CHelpers.CFormat.Float5digit(value_);
            LblDevDatRangeUnitLo.Text = CHartUtils.CHartUnit.GetText(unit_);
            TxtDevDatLowerRange.Refresh();
            LblDevDatRangeUnitLo.Refresh();
        }

        internal void DevDatShowSpan(float value_, byte unit_)
        {
            TxtDevDatSpan.Text = CHelpers.CFormat.Float5digit(value_);
            LblDevDatSpanUnit.Text = CHartUtils.CHartUnit.GetText(unit_);
            TxtDevDatSpan.Refresh();
            LblDevDatSpanUnit.Refresh();
        }
        internal void DevDatShowCurrent(float value_)
        {
            TxtDevDatCurrent.Text = CHelpers.CFormat.Float5digit(value_);
            TxtDevDatCurrent.Refresh();
        }
        internal void DevDatShowPV(float value_, byte unit_)
        {
            TxtDevDatPV.Text = CHelpers.CFormat.Float5digit(value_);
            LblDevDatUnitPV.Text = CHartUtils.CHartUnit.GetText(unit_);
            TxtDevDatPV.Refresh();
            LblDevDatUnitPV.Refresh();
        }
        internal void DevDatShowSV(float value_, byte unit_)
        {
            TxtDevDatSV.Text = CHelpers.CFormat.Float5digit(value_);
            LblDevDatUnitSV.Text = CHartUtils.CHartUnit.GetText(unit_);
            TxtDevDatSV.Refresh();
            LblDevDatUnitSV.Refresh();
        }
        internal void DevDatShowTV(float value_, byte unit_)
        {
            TxtDevDatTV.Text = CHelpers.CFormat.Float5digit(value_);
            LblDevDatUnitTV.Text = CHartUtils.CHartUnit.GetText(unit_);
            TxtDevDatTV.Refresh();
            LblDevDatUnitTV.Refresh();
        }
        internal void DevDatShowQV(float value_, byte unit_)
        {
            TxtDevDatQV.Text = CHelpers.CFormat.Float5digit(value_);
            LblDevDatUnitQV.Text = CHartUtils.CHartUnit.GetText(unit_);
            TxtDevDatQV.Refresh();
            LblDevDatUnitQV.Refresh();
        }
        #endregion Slave Device Data Handling

        #region Cyclic handlers
        internal void HandleWatchingData()
        {
            HandleMonitorData();

            if (CTestClient.hServiceConnect != HartMasterDLL.INVALID_SRV_Handle)
            {
                if (HartMasterDLL.BAHAMA_IsServiceCompleted(CTestClient.hServiceConnect) == EN_Bool.TRUE8)
                {
                    CTestClient.HandleConnectService();
                    CTestClient.hServiceConnect = HartMasterDLL.INVALID_SRV_Handle;
                    CMAStatus.CMonitor.CfgChanged = true;
                    CMAStatus.CGeneral.CfgChanged = true;
                    CMAStatus.CfgChanged = true;
                    HandleControlEvent();
                }

                return;
            }

            if (CTestClient.hServiceDoCmd != HartMasterDLL.INVALID_SRV_Handle)
            {
                if (HartMasterDLL.BAHAMA_IsServiceCompleted(CTestClient.hServiceDoCmd) == EN_Bool.TRUE8)
                {
                    CTestClient.HandleDoCmdService();
                    CTestClient.hServiceDoCmd = HartMasterDLL.INVALID_SRV_Handle;
                }
            }
        }

        internal void HandleWatchingStatus()
        {
            HandleMsgLedStatus();
            if (m_last_number_frames != m_number_frames)
            {
                StatusInfo.Items[3].Text = m_number_frames.ToString("0000000");
                m_last_number_frames = m_number_frames;
            }

            DisplayDiagnostics();
        }

        internal void HandleCyclics()
        {
            HandleWatchingData();
            HandleWatchingStatus();
        }
        #endregion Cyclic handlers

        #region White Test Support
        private void TestBegin(string text_)
        {
            rtxtTestDisplay.Clear();
            rtxtTestDisplay.Text = text_;
            DisableTimers();
        }

        private void DisableTimers()
        {
            TimWatch.Enabled = false;
            TimOperating.Enabled = false;
        }

        private void EnableTimers()
        {
            TimWatch.Enabled = true;
            TimOperating.Enabled = true;
        }

        private void TestUpdate(string text_)
        {
            HandleCyclics();
            rtxtTestDisplay.AppendText("\n" + text_);
        }

        private void TestEnd(string text_)
        {
            rtxtTestDisplay.AppendText("\n" + text_);
            EnableTimers();
        }

        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 256)]
        // Test array buffeer
        private byte[] buffer = new byte[256];

        private void ButUserTestCoding_Click(object sender, EventArgs e)
        {
            for (int i = 0; i < 256; i++)
            {
                buffer[i] = 0;
            }

            byte by = 0;
            TestBegin("User Test Coding Started.");
            TestUpdate("01: Int8(0x11) at 05");
            HartMasterDLL.BAHAMA_PutInt8(0x11, 5, ref buffer[0]);
            by = HartMasterDLL.BAHAMA_PickInt8(5, ref buffer[0]);
            if (by != 0x11)
            {
                TestEnd("Test failed, step 01.");
                return;
            }

            ushort us = 0;
            TestUpdate("02: Int16(0x55AA) at 16");
            HartMasterDLL.BAHAMA_PutInt16(0x55AA, 16, ref buffer[0], EN_Endian.MSB_First);
            us = HartMasterDLL.BAHAMA_PickInt16(16, ref buffer[0], EN_Endian.LSB_First);
            if (us != 0xAA55)
            {
                TestEnd("Test failed, step 02.");
                return;
            }

            uint ui = 0;
            TestUpdate("03: Int16(0x3355AACC) at 88");
            HartMasterDLL.BAHAMA_PutInt32(0x3355AACC, 88, ref buffer[0], EN_Endian.LSB_First);
            ui = HartMasterDLL.BAHAMA_PickInt32(88, ref buffer[0], EN_Endian.MSB_First);
            if (ui != 0xCCAA5533)
            {
                TestEnd("Test failed, step 03.");
                return;
            }

            TestUpdate("04: Int16(0xCCAA) at 88?");
            us = HartMasterDLL.BAHAMA_PickInt16(88, ref buffer[0], EN_Endian.MSB_First);
            if (us != 0xCCAA)
            {
                TestEnd("Test failed, step 04.");
                return;
            }

            TestUpdate("05: Int16(0x5533) at 90?");
            us = HartMasterDLL.BAHAMA_PickInt16(90, ref buffer[0], EN_Endian.MSB_First);
            if (us != 0x5533)
            {
                TestEnd("Test failed, step 05.");
                return;
            }

            TestUpdate("06: PASCII: \"HELLO \"");
            StringBuilder sb = new StringBuilder(8);
            sb.Insert(0, "HELLO ");
            HartMasterDLL.BAHAMA_PutPackedASCII(sb, 6, 30, ref buffer[0]);
            HartMasterDLL.BAHAMA_PickPackedASCII(sb, 6, 30, ref buffer[0]);
            string result = sb.ToString();
            if (result != "HELLO ")
            {
                TestEnd("Test failed, step 06.");
                return;
            }

            TestEnd("Test completed with no errors.");

            // Test sequence still to be completed
        }
        private void ButUserTestCommands_Click(object sender, EventArgs e)
        {
            ushort h_service;
            HartMasterDLL.TY_Confirmation confirmation = new HartMasterDLL.TY_Confirmation();

            for (int i = 0; i < 256; i++)
            {
                buffer[i] = 0;
            }

            TestBegin("User Test Commands Started.");
            TestUpdate("01: Is device connected?");
            CTestClient.ReqCmd0short();
            while (HartMasterDLL.BAHAMA_IsServiceCompleted(CTestClient.hServiceConnect) != EN_Bool.TRUE8)
            {
                HandleCyclics();
                Thread.Sleep(50);
            }

            HandleCyclics();

            if (!CTestClient.IsConnected)
            {
                TestEnd("No device connected.");
                return;
            }

            // Set range to 0 to 100 %
            buffer[0] = 57;
            HartMasterDLL.BAHAMA_PutFloat(100.0f, 1, ref buffer[0], EN_Endian.MSB_First);
            HartMasterDLL.BAHAMA_PutFloat(0.0f, 6, ref buffer[0], EN_Endian.MSB_First);
            TestUpdate("02: Set 0 .. 100 %");
            h_service = HartMasterDLL.BAHAMA_DoCommand(35, HartMasterDLL.WAIT_For_Service, ref buffer[0], 9, ref CTestClient.BytesOfUniqueId[0]);
            HandleCyclics();
            HartMasterDLL.BAHAMA_FetchConfirmation(h_service, ref confirmation);
            if (confirmation.RespCode1 != 0)
            {
                TestEnd("Command rejected.");
                return;
            }

            // Read back the range values
            TestUpdate("03: Read back range");
            h_service = HartMasterDLL.BAHAMA_DoCommand(15, HartMasterDLL.WAIT_For_Service, ref buffer[0], 0, ref CTestClient.BytesOfUniqueId[0]);
            HandleCyclics();
            HartMasterDLL.BAHAMA_FetchConfirmation(h_service, ref confirmation);
            if (confirmation.RespCode1 != 0)
            {
                TestEnd("Command rejected.");
                return;
            }

            float upper_range = HartMasterDLL.BAHAMA_PickFloat(3, ref confirmation.BytesOfData[0], EN_Endian.MSB_First);
            if (upper_range != 100.0f)
            {
                TestEnd("Upper range error!");
                return;
            }

            float lower_range = HartMasterDLL.BAHAMA_PickFloat(7, ref confirmation.BytesOfData[0], EN_Endian.MSB_First);
            if (lower_range != 0.0f)
            {
                TestEnd("Lower range error!");
                return;
            }

            TestEnd("Test completed with no errors.");
        }

        private void ButFreeTest_Click(object sender, EventArgs e)
        {
            TestBegin("Free User Test Started.");
            TestUpdate("No test sequence has yet been ");
            TestUpdate("implemented for this button.");
            TestEnd("Test completed with no errors.");
        }
        #endregion White Test Support
    }
}
